#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""DICOM fields for attribute serialization."""

import datetime
import math
from abc import abstractmethod
from typing import (
    Any,
    Dict,
    Generic,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from marshmallow import Schema, fields
from marshmallow.utils import missing
from pydicom import DataElement, Dataset
from pydicom.multival import MultiValue
from pydicom.sr.coding import Code
from pydicom.uid import UID
from pydicom.valuerep import DA, DT, TM, DSfloat, PersonName

from wsidicom.conceptcode import ConceptCode, UnitCode
from wsidicom.geometry import Orientation, PointMm, SizeMm
from wsidicom.metadata.sample import (
    IssuerOfIdentifier,
    LocalIssuerOfIdentifier,
    Measurement,
    UniversalIssuerOfIdentifier,
    UniversalIssuerType,
)


class StringLikeDicomField(fields.Field):
    def _deserialize(self, value, attr, data, **kwargs) -> Any:
        """By default pydicom returns empty string for empty string-like elements."""
        deserialized = super()._deserialize(value, attr, data, **kwargs)
        if deserialized == "":
            return None
        return deserialized


class StringDicomField(StringLikeDicomField):
    pass


class EnumDicomField(fields.Enum):
    def _deserialize(self, value, attr, data, **kwargs):
        if value == "":
            return None
        return super()._deserialize(value, attr, data, **kwargs)


class DateTimeDicomField(StringLikeDicomField):
    def _serialize(
        self,
        value: Optional[datetime.datetime],
        attr: Optional[str],
        obj: Any,
        **kwargs,
    ):
        if value is None:
            return None
        return DT(value)

    def _deserialize(self, value, attr, data, **kwargs) -> Any:
        if value is None:
            return None
        try:
            return DT(value)
        except ValueError:
            return None


class DateDicomField(StringLikeDicomField):
    def _serialize(
        self, value: Optional[datetime.date], attr: Optional[str], obj: Any, **kwargs
    ):
        if value is None:
            return None
        return DA(value)

    def _deserialize(self, value, attr, data, **kwargs) -> Any:
        if value is None:
            return None
        try:
            return DA(value)
        except ValueError:
            return None


class TimeDicomField(StringLikeDicomField):
    def _serialize(
        self, value: Optional[datetime.time], attr: Optional[str], obj: Any, **kwargs
    ):
        if value is None:
            return None
        return TM(value)

    def _deserialize(self, value, attr, data, **kwargs) -> Any:
        if value is None:
            return None
        try:
            return TM(value)
        except ValueError:
            return None


class BooleanDicomField(fields.Boolean):
    def __init__(
        self,
        truthy: Optional[str] = None,
        falsy: Optional[str] = None,
        **kwargs,
    ):
        if truthy is None:
            truthy = "YES"
        if falsy is None:
            falsy = "NO"
        super().__init__(truthy=set([truthy]), falsy=set([falsy]), **kwargs)

    def _serialize(self, value: bool, attr: Optional[str], obj: Any, **kwargs):
        if value:
            string_value = self.truthy
        else:
            string_value = self.falsy
        return list(string_value)[0]

    def _deserialize(self, value, attr, data, **kwargs) -> Any:
        deserialized = super()._deserialize(value, attr, data, **kwargs)
        if deserialized == "":
            return None
        return deserialized


class OffsetInSlideCoordinateSystemField(fields.Field):
    def _serialize(
        self, origin: Optional[PointMm], attr: Optional[str], obj: Any, **kwargs
    ):
        if origin is None:
            if self.dump_default is None:
                return None
            assert isinstance(self.dump_default, PointMm)
            origin = self.dump_default
        origin_element = Dataset()
        origin_element.XOffsetInSlideCoordinateSystem = DSfloat(origin.x, True)
        origin_element.YOffsetInSlideCoordinateSystem = DSfloat(origin.y, True)
        return [origin_element]

    def _deserialize(
        self,
        value: DataElement,
        attr: Optional[str],
        data: Optional[Dict[str, Any]],
        **kwargs,
    ) -> PointMm:
        return PointMm(
            x=value[0].XOffsetInSlideCoordinateSystem,
            y=value[0].YOffsetInSlideCoordinateSystem,
        )


class ImageOrientationSlideField(fields.Field):
    def _serialize(
        self, rotation: Optional[float], attr: Optional[str], obj: Any, **kwargs
    ):
        if rotation is None:
            if self.dump_default is None:
                return None
            assert isinstance(self.dump_default, float)
            rotation = self.dump_default
        x = round(math.sin(rotation * math.pi / 180), 8)
        y = round(math.cos(rotation * math.pi / 180), 8)
        return [-x, y, 0, y, x, 0]

    def _deserialize(
        self,
        value: Tuple[float, float, float, float, float, float],
        attr: Optional[str],
        data: Optional[Dict[str, Any]],
        **kwargs,
    ) -> float:
        orientation = Orientation(value)
        return orientation.rotation


class ListDicomField(fields.List):
    """Wrapper around normal list that handles single-valued lists from pydicom."""

    def __init__(
        self,
        cls_or_instance: Union[fields.Field, type],
        dump_none_if_empty: bool = False,
        **kwargs,
    ):
        self._dump_none_if_empty = dump_none_if_empty
        super().__init__(cls_or_instance, **kwargs)

    def _serialize(self, value, attr, obj, **kwargs) -> Any:
        if self._dump_none_if_empty and value is None or len(value) == 0:
            return None
        return super()._serialize(value, attr, obj, **kwargs)

    def _deserialize(
        self, value: Union[Any, List[Any]], attr, data, **kwargs
    ) -> List[Any]:
        if not isinstance(value, MultiValue):
            value = [value]
        return super()._deserialize(value, attr, data, **kwargs)


class FlattenOnDumpNestedDicomField(fields.Nested):
    """Field that flattens the nested dataset into the parent dataset on dump.

    On load the nested fields are deflatten from the parent dataset to a nested dataset.

    The flatten/deflatten is done by the parent schema.
    """

    def __init__(self, nested: Schema, **kwargs):
        self._nested = nested
        super().__init__(nested=nested, **kwargs)

    @property
    def nested_schema(self) -> Schema:
        return self._nested

    def de_flatten(self, dataset: Dataset) -> Optional[Dataset]:
        """Create new dataset containing the attributes defined in nested schema."""
        nested = Dataset()
        for nested_field in self.nested_schema.fields.values():
            if nested_field.dump_only:
                continue
            if isinstance(nested_field, FlattenOnDumpNestedDicomField):
                de_flatten_nested_field = nested_field.de_flatten(dataset)
                if de_flatten_nested_field is not None:
                    nested.update(de_flatten_nested_field)
            elif nested_field.data_key is not None and nested_field.data_key in dataset:
                nested_value = dataset.get(nested_field.data_key)
                setattr(nested, nested_field.data_key, nested_value)
        if len(nested) == 0:
            return None
        return nested

    def flatten(self, data: Dict[str, Any]):
        """Insert attributes from nested dataset into data."""
        key = self.name
        if self.data_key is not None:
            key = self.data_key
        nested = data.pop(key, None)
        if isinstance(nested, Dataset):
            for nested_key, nested_value in nested.items():
                data[nested_key] = nested_value  # type: ignore

    def _serialize(self, nested_obj, attr: Optional[str], obj: Any, **kwargs):
        if nested_obj is None and self.dump_default != missing:
            nested_obj = self.dump_default
        return super()._serialize(nested_obj, attr, obj, **kwargs)


class FlattenOnLoadNestedDicomField(fields.Nested):
    """Field that flattens the nested loaded item into the parent item on load.

    On dump the nested fields are deflatten from the parent dataset to a nested dataset.
    """

    def __init__(self, nested: Schema, **kwargs):
        self._nested = nested
        super().__init__(nested=nested, **kwargs)

    @property
    def nested_schema(self) -> Schema:
        return self._nested

    def de_flatten(self, dataset: Dataset) -> Optional[Dataset]:
        """Create new dataset containing the attributes defined in nested schema."""
        nested = Dataset()
        for nested_field in self.nested_schema.fields.values():
            if nested_field.dump_only:
                continue
            if isinstance(nested_field, FlattenOnDumpNestedDicomField):
                de_flatten_nested_field = nested_field.de_flatten(dataset)
                if de_flatten_nested_field is not None:
                    nested.update(de_flatten_nested_field)
            elif nested_field.data_key is not None and nested_field.data_key in dataset:
                nested_value = dataset.get(nested_field.data_key)
                setattr(nested, nested_field.data_key, nested_value)
        if len(nested) == 0:
            return None
        return nested

    def flatten(self, data: Dict[str, Any]):
        """Insert attributes from nested dataset into data."""
        key = self.name
        if self.data_key is not None:
            key = self.data_key
        nested = data.pop(key, None)
        if isinstance(nested, Dataset):
            for nested_key, nested_value in nested.items():
                data[nested_key] = nested_value  # type: ignore


CodeType = TypeVar("CodeType", Code, ConceptCode)


class FloatDicomField(fields.Float):
    def _serialize(self, value: float, attr: Optional[str], obj: Any, **kwargs):
        return DSfloat(value)


class CodeDicomField(fields.Field, Generic[CodeType]):
    def __init__(self, load_type: Type[CodeType], **kwargs) -> None:
        self._load_type = load_type
        super().__init__(**kwargs)

    def _serialize(
        self, value: Optional[CodeType], attr: Optional[str], obj: Any, **kwargs
    ):
        if value is None:
            return self.dump_default
        dataset = Dataset()
        dataset.CodeValue = value.value
        dataset.CodingSchemeDesignator = value.scheme_designator
        dataset.CodeMeaning = value.meaning
        if value.scheme_version is not None and value.scheme_version != "":
            dataset.CodingSchemeVersion = value.scheme_version

        return dataset

    def _deserialize(
        self,
        value: Dataset,
        attr: Optional[str],
        data: Optional[Dict[str, Any]],
        **kwargs,
    ):
        version = value.get("CodingSchemeVersion", None)
        if version == "":
            version = None
        return self._load_type(
            value=value.CodeValue,
            scheme_designator=value.CodingSchemeDesignator,
            meaning=value.CodeMeaning,
            scheme_version=version,
        )


class SingleCodeSequenceField(CodeDicomField):
    """Field for a DICOM `code sequence` that can only contain one code."""

    def _serialize(self, value: CodeType, attr: Optional[str], obj: Any, **kwargs):
        return [super()._serialize(value, attr, obj, **kwargs)]

    def _deserialize(
        self,
        value: Sequence[Dataset],
        attr: Optional[str],
        data: Optional[Dict[str, Any]],
        **kwargs,
    ):
        return super()._deserialize(value[0], attr, data, **kwargs)


class FloatOrCodeDicomField(fields.Field, Generic[CodeType]):
    def __init__(self, load_type: Type[CodeType], **kwargs) -> None:
        self._float_field = FloatDicomField()
        self._code_field = CodeDicomField(load_type)
        super().__init__(**kwargs)

    def _serialize(
        self, value: Union[float, CodeType], attr: Optional[str], obj: Any, **kwargs
    ):
        assert attr is not None
        if isinstance(value, float):
            return self._float_field.serialize(attr, obj, **kwargs)
        return self._code_field.serialize(attr, obj, **kwargs)

    def _deserialize(self, value: Union[DSfloat, Dataset], attr, data, **kwargs):
        if isinstance(value, DSfloat):
            return self._float_field.deserialize(value, attr, data, **kwargs)
        return self._code_field.deserialize(value, attr, data, **kwargs)


class UidDicomField(fields.Field):
    def _deserialize(self, value: Any, attr, data, **kwargs):
        if value is None or value == "":
            return None
        if isinstance(value, UID):
            return value
        return UID(value)


class UidDatasetDicomField(UidDicomField):
    def __init__(self, data_key: str, **kwargs):
        self._data_key = data_key
        super().__init__(data_key=data_key, **kwargs)

    def _serialize(
        self, value: Optional[UID], attr: Optional[str], obj: Any, **kwargs
    ) -> Optional[Dataset]:
        if value is None:
            return None
        dataset = Dataset()
        setattr(dataset, self._data_key, value)
        return dataset

    def _deserialize(self, value: Dataset, attr, data, **kwargs) -> Optional[UID]:
        nested_value = getattr(value, self._data_key, None)
        if nested_value is None:
            return None
        return super()._deserialize(nested_value, attr, data, **kwargs)


class PatientNameDicomField(fields.String):
    def _deserialize(self, value: PersonName, attr, data, **kwargs) -> str:
        return str(value)

    def _serialize(self, value: str, attr, obj, **kwargs) -> Optional[PersonName]:
        return PersonName(value)


class IssuerOfIdentifierDicomField(fields.Field):
    def _deserialize(
        self, value: Optional[Sequence[Dataset]], attr, data, **kwargs
    ) -> Optional[IssuerOfIdentifier]:
        if value is None or len(value) == 0:
            return None
        dataset = value[0]
        if "UniversalEntityIDType" in dataset:
            return UniversalIssuerOfIdentifier(
                dataset.UniversalEntityID,
                UniversalIssuerType(dataset.UniversalEntityIDType),
                dataset.get("LocalNamespaceEntityID", None),
            )
        if "LocalNamespaceEntityID" in dataset:
            return LocalIssuerOfIdentifier(dataset.LocalNamespaceEntityID)
        return None

    def _serialize(
        self, value: Optional[IssuerOfIdentifier], attr, obj, **kwargs
    ) -> Optional[Sequence[Dataset]]:
        if value is None:
            return []
        dataset = Dataset()
        if isinstance(value, UniversalIssuerOfIdentifier):
            dataset.UniversalEntityID = value.identifier
            dataset.UniversalEntityIDType = value.issuer_type.name
            if value.local_identifier is not None:
                dataset.LocalNamespaceEntityID = value.local_identifier
        elif isinstance(value, LocalIssuerOfIdentifier):
            dataset.LocalNamespaceEntityID = value.identifier
        else:
            raise NotImplementedError()
        return [dataset]


class PixelSpacingDicomField(fields.Field):
    def _serialize(
        self, value: Optional[SizeMm], attr: Optional[str], obj: Any, **kwargs
    ):
        if value is None:
            return None
        return [DSfloat(value.width, True), DSfloat(value.height, True)]

    def _deserialize(
        self, value: Optional[Sequence[DSfloat]], attr, data, **kwargs
    ) -> Optional[SizeMm]:
        if value is None or len(value) == 0:
            return None
        return SizeMm(value[0], value[1])


ValueType = TypeVar("ValueType")


class TypeDicomField(fields.Field, Generic[ValueType]):
    def __init__(self, nested: fields.Field, **kwargs):
        self._nested = nested
        super().__init__(**kwargs)

    def _serialize(
        self, value: Optional[ValueType], attr: Optional[str], obj: Any, **kwargs
    ):
        return self._nested._serialize(value, attr, obj, **kwargs)

    def _deserialize(self, value: Any, attr, data, **kwargs):
        de_serialized = self._nested._deserialize(value, attr, data, **kwargs)
        return de_serialized


class DefaultingDicomField(TypeDicomField[ValueType]):
    """Wrapper around a field that should always be present and have a value. Default
    value is constant."""

    def __init__(self, nested: fields.Field, dump_default: ValueType, **kwargs):
        self._dump_default = dump_default
        super().__init__(nested=nested, dump_default=dump_default, **kwargs)

    def _serialize(
        self, value: Optional[ValueType], attr: Optional[str], obj: Any, **kwargs
    ):
        if value is None:
            value = self._dump_default
        return super()._serialize(value, attr, obj, **kwargs)


class DefaultingNoneDicomField(DefaultingDicomField[Optional[ValueType]]):
    """Wrapper around a field that should always be present but can be None."""

    def __init__(self, nested: fields.Field, **kwargs):
        super().__init__(nested=nested, dump_default=None, **kwargs)

    def _serialize(
        self, value: Optional[ValueType], attr: Optional[str], obj: Any, **kwargs
    ):
        return super()._serialize(value, attr, obj, **kwargs)


class DefaultingTagDicomField(TypeDicomField[ValueType]):
    """Wrapper around a field that should always be present and have a value. Default
    value is taken from object by attribute defined by tag."""

    def __init__(self, nested: fields.Field, tag: str, **kwargs):
        self._tag = tag
        super().__init__(nested=nested, **kwargs)

    def _serialize(
        self, value: Optional[ValueType], attr: Optional[str], obj: Any, **kwargs
    ):
        if value is None:
            value = getattr(obj, self._tag)
        return super()._serialize(value, attr, obj, **kwargs)


class DefaultingListDicomField(fields.List):
    def __init__(self, nested: fields.Field, dump_default: List, **kwargs):
        self._dump_default = dump_default
        super().__init__(cls_or_instance=nested, dump_default=dump_default, **kwargs)

    def _serialize(self, value, attr, obj, **kwargs) -> Any:
        if value is None or len(value) == 0:
            value = self.dump_default
        return super()._serialize(value, attr, obj, **kwargs)


class DefaultingListTagDicomField(fields.List):
    def __init__(self, nested: fields.Field, tag: str, **kwargs):
        self._tag = tag
        super().__init__(cls_or_instance=nested, **kwargs)

    def _serialize(self, value, attr, obj, **kwargs) -> Any:
        if value is None or len(value) == 0:
            value = getattr(obj, self._tag)
        return super()._serialize(value, attr, obj, **kwargs)


class NestedDatasetDicomField(fields.Nested, Generic[ValueType]):
    """fields.Field for attribute of a single-item dataset sequence with a nested
    sing-item dataset sequence with the item the nested schema should handle."""

    def __init__(self, nested: Schema, data_key: str, nested_data_key: str, **kwargs):
        self._nested = nested
        self._data_key = data_key
        self._nested_data_key = nested_data_key
        super().__init__(nested=nested, data_key=data_key, **kwargs)

    def _serialize(
        self, value: Optional[ValueType], attr: Optional[str], obj: Any, **kwargs
    ) -> Optional[List[Dataset]]:
        nested_value = super()._serialize(value, attr, obj, **kwargs)
        if nested_value is None:
            return None
        dataset = Dataset()
        setattr(dataset, self._nested_data_key, [nested_value])
        return [dataset]

    def _deserialize(
        self, value: Sequence[Dataset], attr, data, **kwargs
    ) -> Optional[ValueType]:
        nested_value = getattr(value[0], self._nested_data_key, None)
        if nested_value is None:
            return None
        return super()._deserialize(nested_value[0], attr, data, **kwargs)  # type: ignore


class ContentItemDicomField(fields.Field, Generic[ValueType]):
    @abstractmethod
    def _serialize(
        self, value: Optional[ValueType], attr: Optional[str], obj: Any, **kwargs
    ) -> Optional[Dataset]:
        raise NotImplementedError()

    @abstractmethod
    def _deserialize(
        self, value: Dataset, attr: Optional[str], obj: Any, **kwargs
    ) -> Optional[ValueType]:
        raise NotImplementedError()


class CodeItemDicomField(ContentItemDicomField[Code]):
    def __init__(self, load_type: Type[CodeType], **kwargs) -> None:
        self._load_type = load_type
        super().__init__(**kwargs)

    def _serialize(
        self, value: Optional[Code], attr: Optional[str], obj: Any, **kwargs
    ) -> Optional[Dataset]:
        if value is None:
            return None
        code_dataset = Dataset()
        code_dataset.CodeValue = value.value
        code_dataset.CodingSchemeDesignator = value.scheme_designator
        code_dataset.CodeMeaning = value.meaning
        if value.scheme_version is not None and value.scheme_version != "":
            code_dataset.CodingSchemeVersion = value.scheme_version
        dataset = Dataset()
        dataset.ValueType = "CODE"
        dataset.ConceptCodeSequence = [code_dataset]
        return dataset

    def _deserialize(self, dataset: Dataset, attr: Optional[str], obj: Any, **kwargs):
        version = dataset.ConceptCodeSequence[0].get("CodingSchemeVersion", None)
        if version == "":
            version = None
        return self._load_type(
            value=dataset.ConceptCodeSequence[0].CodeValue,
            scheme_designator=dataset.ConceptCodeSequence[0].CodingSchemeDesignator,
            meaning=dataset.ConceptCodeSequence[0].CodeMeaning,
            scheme_version=version,
        )


class StringItemDicomField(ContentItemDicomField[str]):
    def _serialize(
        self, value: Optional[str], attr: Optional[str], obj: Any, **kwargs
    ) -> Optional[Dataset]:
        if value is None:
            return None
        dataset = Dataset()
        dataset.ValueType = "TEXT"
        dataset.TextValue = value
        return dataset

    def _deserialize(self, dataset: Dataset, attr: Optional[str], obj: Any, **kwargs):
        return dataset.TextValue


class StringOrCodeItemDicomField(ContentItemDicomField[Union[str, Code]]):
    def __init__(self, load_type: Type[CodeType], **kwargs) -> None:
        self._string_field = StringItemDicomField()
        self._code_field = CodeItemDicomField(load_type)
        super().__init__(**kwargs)

    def _serialize(
        self, value: Optional[Union[str, Code]], attr: str, obj: Any, **kwargs
    ) -> Optional[Dataset]:
        if value is None:
            return None
        if isinstance(value, str):
            return self._string_field._serialize(value, attr, obj, **kwargs)
        return self._code_field._serialize(value, attr, obj, **kwargs)

    def _deserialize(self, dataset: Dataset, attr: str, obj: Any, **kwargs):
        if dataset.ValueType == "TEXT" or hasattr(dataset, "TextValue"):
            return self._string_field.deserialize(dataset, attr, obj, **kwargs)
        return self._code_field.deserialize(dataset, attr, obj, **kwargs)


class DateTimeItemDicomField(ContentItemDicomField[datetime.datetime]):
    def _serialize(
        self,
        value: Optional[datetime.datetime],
        attr: Optional[str],
        obj: Any,
        **kwargs,
    ) -> Optional[Dataset]:
        if value is None:
            return None
        dataset = Dataset()
        dataset.ValueType = "DATETIME"
        dataset.DateTime = DT(value)
        return dataset

    def _deserialize(self, dataset: Dataset, attr: Optional[str], obj: Any, **kwargs):
        try:
            return DT(dataset.DateTime)
        except ValueError:
            return None


class MeasurementtemDicomField(ContentItemDicomField[Measurement]):
    def _serialize(
        self, value: Optional[Measurement], attr: Optional[str], obj: Any, **kwargs
    ) -> Optional[Dataset]:
        if value is None:
            return None
        dataset = Dataset()
        dataset.ValueType = "NUMERIC"
        dataset.NumericValue = DSfloat(value.value)
        dataset.FloatingPointValue = value.value
        unit_dataset = Dataset()
        unit_dataset.CodeValue = value.unit.value
        unit_dataset.CodingSchemeDesignator = value.unit.scheme_designator
        unit_dataset.CodeMeaning = value.unit.meaning
        if value.unit.scheme_version is not None and value.unit.scheme_version != "":
            unit_dataset.CodingSchemeVersion = value.unit.scheme_version
        dataset.MeasurementUnitsCodeSequence = [unit_dataset]
        return dataset

    def _deserialize(self, dataset: Dataset, attr: Optional[str], obj: Any, **kwargs):
        if hasattr(dataset, "FloatingPointValue"):
            value = dataset.FloatingPointValue
        else:
            value = DSfloat(dataset.NumericValue)
            assert isinstance(value, float)
        unit_dataset = dataset.MeasurementUnitsCodeSequence[0]
        version = unit_dataset.get("CodingSchemeVersion", None)
        if version == "":
            version = None
        unit = UnitCode(
            value=unit_dataset.CodeValue,
            scheme_designator=unit_dataset.CodingSchemeDesignator,
            meaning=unit_dataset.CodeMeaning,
            scheme_version=version,
        )
        return Measurement(value=value, unit=unit)



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from abc import ABCMeta, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import (
    Any,
    BinaryIO,
    Callable,
    DefaultDict,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import numpy as np
from pydicom.dataset import Dataset, FileDataset, FileMetaDataset, validate_file_meta
from pydicom.filereader import dcmread
from pydicom.filewriter import dcmwrite
from pydicom.sequence import Sequence as DicomSequence
from pydicom.uid import (
    ExplicitVRBigEndian,
    ExplicitVRLittleEndian,
    ImplicitVRLittleEndian,
    generate_uid,
)
from pydicom.values import convert_numbers

from wsidicom.conceptcode import (
    AnnotationCategoryCode,
    AnnotationTypeCode,
    MeasurementCode,
    UnitCode,
)

from .geometry import PointMm, RegionMm, SizeMm
from .uid import ANN_SOP_CLASS_UID, UID, SlideUids


@dataclass(frozen=True)
class LabColor:
    l: int
    a: int
    b: int


def dcm_to_list(item: bytes, format: str) -> List[Any]:
    """Convert item to list of values using specified type format.

    Parameters
    ----------
    item: bytes
        Bytes to convert.
    format: str
        Struct format character.

    Returns
    ----------
    List[Any]
        List of values
    """
    converted = convert_numbers(item, is_little_endian=True, struct_format=format)
    if not isinstance(converted, list):
        return [converted]
    return converted


@dataclass
class Measurement:
    """Represents a measurement.

    Parameters
        ----------
        code: MeasurementCode
            Type of measurement.
        value: float
            Value of measurement.
        unit: UnitCode
            Unit of measurement.
    """

    code: MeasurementCode
    value: float
    unit: UnitCode

    def same_type(self, other_code: MeasurementCode, other_unit: UnitCode):
        """Return true if measurement base is same.

        Parameters
        ----------
        other_code: MeasurementCode
            Other base to compare to.
        other_unit: UnitCode
            Other base to compare to.

        Returns
        -------
        bool
            True if bases are same.
        """
        return self.code == other_code and self.unit == other_unit

    @staticmethod
    def _get_measurement_type_from_ds(ds: Dataset) -> Tuple[MeasurementCode, UnitCode]:
        """Get measurement type from dataset.

        Parameters
        ----------
        ds: Dataset
            Dataset containing measurement sequence item (group).

        Returns
        -------
        MeasurementType
            Measurement type of the measurement group.
        """
        code = MeasurementCode.from_ds(ds)
        if code is None:
            raise ValueError(f"Dataset is missing {MeasurementCode.sequence_name}.")
        unit = UnitCode.from_ds(ds)
        if unit is None:
            raise ValueError(f"Dataset is missing {UnitCode.sequence_name}.")
        return code, unit

    @staticmethod
    def _get_measurement_values_from_ds(ds: Dataset) -> List[float]:
        """Get measurement values from dataset.

        Parameters
        ----------
        ds: Dataset
            Dataset containing measurement sequence item (group).

        Returns
        -------
        List[float]
            Measurement values of the measurement group.
        """
        return dcm_to_list(ds.MeasurementValuesSequence[0].FloatingPointValues, "f")

    @staticmethod
    def _get_measurement_indices_from_ds(
        ds: Dataset,
        annotation_count: int,
    ) -> List[int]:
        """Get measurement indices from dataset. Annotation index is stored as
        starting at index 1. If Annotation index list is
        missing measurements are one-to-one to annotations.

        Parameters
        ----------
        ds: Dataset
            Dataset containing measurement sequence item (group).
        annotation_count: int
            Number of annotations.

        Returns
        -------
        List[int]
            Measurement indices of the measurement group.
        """
        measurement_ds = ds.MeasurementValuesSequence[0]
        if "AnnotationIndexList" in measurement_ds:
            indices = dcm_to_list(measurement_ds.AnnotationIndexList, "l")
            return [index - 1 for index in indices]
        return list(range(annotation_count))

    @classmethod
    def _iterate_measurement_sequence(
        cls, sequence: DicomSequence, annotation_count: int
    ) -> Iterator[Tuple[int, "Measurement"]]:
        """Return generator for measurements in dataset. Yields a tuple of
        the index of the annotation for the measurement and the measurement.

        Parameters
        ----------
        sequence: DicomSequence
            Dicom measurement sequence.
        annotation_count: int
            Number of annotations in group.

        Returns
        -------
        Iterator[Tuple[int, 'Measurement']]:
            Generator for annotation index and measurement.
        """
        for group_ds in sequence:
            code, unit = cls._get_measurement_type_from_ds(group_ds)
            values = cls._get_measurement_values_from_ds(group_ds)
            indices = cls._get_measurement_indices_from_ds(group_ds, annotation_count)
            if len(values) != len(indices):
                raise ValueError("number of indices needs to be same as measurements")
            for index, annotation_index in enumerate(indices):
                yield (annotation_index, Measurement(code, values[index], unit))

    @classmethod
    def get_measurements_from_ds(
        cls, sequence: DicomSequence, annotation_count: int
    ) -> Dict[int, List["Measurement"]]:
        """Get measurements from dataset.

        Parameters
        ----------
        sequence: DicomSequence
            Dicom measurement sequence.
        annotation_count: int
            Number of annotations in group.

        Returns
        -------
        Dict[int, List[Measurement]]
            Dict of measurements grouped by annotation number as key.
        """
        measurements: DefaultDict[int, List[Measurement]] = defaultdict(list)
        for annotation_index, measurement in cls._iterate_measurement_sequence(
            sequence, annotation_count
        ):
            measurements[annotation_index].append(measurement)
        return measurements


class Geometry(metaclass=ABCMeta):
    name: str

    @property
    @abstractmethod
    def data(self) -> List[float]:
        """Return geometry content as a list of floats"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def box(self) -> RegionMm:
        """Return Region that contains the geometry."""
        raise NotImplementedError()

    @abstractmethod
    def to_coords(self) -> List[Tuple[float, float]]:
        """Return geometry content as a list of tuple of floats"""
        raise NotImplementedError()

    @abstractmethod
    def to_list_coords(self) -> List[List[float]]:
        raise NotImplementedError()

    @abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def from_list(cls, list: Sequence[float]) -> "Geometry":
        """Return geometry object created from list of floats"""
        raise NotImplementedError()

    @classmethod
    def list_to_coords(cls, data: Sequence[float]) -> List[Tuple[float, float]]:
        """Return coordinates in list of floats as list of tuple of floats

        Parameters
        ----------
        data: Sequence[float]
            List of float to convert.

        Returns
        -------
        List[Tuple[float, float]]
            List of coordinates (Tuple of floats).
        """
        x_indices = range(0, len(data), 2)
        y_indices = range(1, len(data), 2)
        return [(float(data[x]), float(data[y])) for x, y in zip(x_indices, y_indices)]

    @staticmethod
    def _coordinates_from_dict(
        dictionary: Union[
            Dict[str, Union[str, float]], Sequence[Dict[str, Union[str, float]]]
        ],
        x: str,
        y: str,
    ) -> List[Tuple[float, float]]:
        """Return coordinates in dictionary as list of tuple of floats

        Parameters
        ----------
        dictionary: Union[
            Dict[str, Union[str, float]],
            Sequence[Dict[str, Union[str, float]]]
            ]:
            Dictionary to convert.
        x: str
            Name of the x coordinate.
        y: str
            Name of the y coordinate.

        Returns
        -------
        List[Tuple[float, float]]
            List of coordinates (Tuple of floats).
        """
        if not isinstance(dictionary, Sequence):
            coordinates = [dictionary]
        else:
            coordinates = dictionary
        return [
            (float(coordinate[x]), float(coordinate[y])) for coordinate in coordinates
        ]

    @classmethod
    @abstractmethod
    def from_dict(
        cls,
        dictionary: Union[
            Dict[str, Union[str, float]], Sequence[Dict[str, Union[str, float]]]
        ],
        x: str,
        y: str,
    ) -> "Geometry":
        raise NotImplementedError()

    @classmethod
    def from_shapely_like(cls, item: Any) -> "Geometry":
        """Return Geometry from shapely-like object. Object needs to have
        shapely-like attributes.

        Parameters
        ----------
        item
            Object with shapely-like attributes.

        Returns
        -------
        Geometry
            Geometry created from object.
        """
        try:
            geometry_type = getattr(item, "geom_type")
            if geometry_type == "Point":
                coordinates = getattr(item, "coords")
                return Point(*coordinates[0])
            elif geometry_type == "Polygon":
                interiors = getattr(item, "interiors")
                exterior = getattr(item, "exterior")
                exterior_coords = getattr(exterior, "coords")
                if len(interiors) == 0:
                    return Polygon(exterior_coords)
            elif geometry_type == "LineString":
                coordinates = getattr(item, "coords")
                return Polyline(coordinates)
            raise NotImplementedError("Not a supported shapely like object")
        except AttributeError as exception:
            raise ValueError("Not a shapely like object") from exception

    @classmethod
    def from_geojson(cls, dictionary: Dict[str, Any]) -> List["Geometry"]:
        """Return geometries from geojson geometry dictionary. Note that
        MultiPoint returns multiple points.

        Parameters
        ----------
        dictionary
            Geojson geometry dictionary.

        Returns
        -------
        Geometry
            Geometry created from Geojson dictionary.
        """
        try:
            coordinates = dictionary["coordinates"]
            annotation_type = dictionary["type"]
        except KeyError:
            raise ValueError("Not a geojson object")
        if not isinstance(coordinates, list) or not isinstance(annotation_type, str):
            raise ValueError("Not a geojson object")
        if annotation_type == "Point":
            return [Point.from_list(coordinates)]
        elif annotation_type == "MultiPoint":
            return [Point.from_list(point) for point in coordinates]
        elif annotation_type == "Polygon":
            return [Polygon(coordinates[0])]
        elif annotation_type == "LineString":
            return [Polyline(coordinates)]
        raise NotImplementedError("Not supported geojson geometry")


@dataclass
class Point(Geometry):
    """Geometry consisting of a single point"""

    name = "POINT"

    def __init__(self, x: float, y: float):
        self.x = float(x)
        self.y = float(y)

    def __eq__(self, point: "Point") -> bool:
        if isinstance(point, Point):
            return self.x == point.x and self.y == point.y
        else:
            raise NotImplementedError("Comparing Point to non-Point")

    @property
    def data(self) -> List[float]:
        return [self.x, self.y]

    @property
    def box(self) -> RegionMm:
        return RegionMm(PointMm(self.x, self.y), SizeMm(0, 0))

    def __str__(self) -> str:
        return f"x: {self.x}, y: {self.y}"

    def __repr__(self) -> str:
        return f"Point({self.x}, {self.y})"

    def to_coords(self) -> List[Tuple[float, float]]:
        return [(self.x, self.y)]

    def to_list_coords(self) -> List[List[float]]:
        return [[self.x, self.y]]

    def __len__(self) -> int:
        return 1

    @classmethod
    def from_list(cls, list: List[float]) -> "Point":
        coordinates = cls.list_to_coords(list)
        return cls(*coordinates[0])

    @classmethod
    def multiple_from_dict(
        cls,
        dictionary: Union[
            Dict[str, Union[str, float]], Sequence[Dict[str, Union[str, float]]]
        ],
        x: str,
        y: str,
    ) -> List["Point"]:
        coordinates = cls._coordinates_from_dict(dictionary, x, y)
        return [cls(*point) for point in coordinates]

    @classmethod
    def from_dict(
        cls,
        dictionary: Union[
            Dict[str, Union[str, float]], Sequence[Dict[str, Union[str, float]]]
        ],
        x: str,
        y: str,
    ) -> "Point":
        coordinates = cls._coordinates_from_dict(dictionary, x, y)
        return cls(*coordinates[0])


@dataclass
class Polyline(Geometry):
    """Geometry consisting of connected lines."""

    name = "POLYLINE"

    def __init__(self, points: Sequence[Tuple[float, float]]):
        self.points: List[Point] = [Point(point[0], point[1]) for point in points]

    def __len__(self) -> int:
        return len(self.points)

    def __repr__(self) -> str:
        return f"Polyline({self.to_coords()})"

    def to_coords(self) -> List[Tuple[float, float]]:
        return [(point.x, point.y) for point in self.points]

    def to_list_coords(self) -> List[List[float]]:
        return [[point.x, point.y] for point in self.points]

    @property
    def data(self) -> List[float]:
        return [value for point in self.points for value in point.data]

    @property
    def box(self) -> RegionMm:
        top: float = self.points[0].y
        left: float = self.points[0].x
        bottom: float = self.points[0].y
        right: float = self.points[0].x
        for point in self.points[1:]:
            if point.y > top:
                top = point.y
            elif point.y < bottom:
                bottom = point.y
            if point.x > right:
                right = point.x
            elif point.y < left:
                left = point.x
        return RegionMm(PointMm(left, bottom), SizeMm(right - left, top - bottom))

    @classmethod
    def from_dict(
        cls,
        dictionary: Union[
            Dict[str, Union[str, float]], Sequence[Dict[str, Union[str, float]]]
        ],
        x: str,
        y: str,
    ) -> "Polyline":
        coordinates = cls._coordinates_from_dict(dictionary, x, y)
        return cls(coordinates)

    @classmethod
    def from_list(cls, list: Sequence[float]) -> "Polyline":
        coordinates = cls.list_to_coords(list)
        return cls(coordinates)


@dataclass
class Polygon(Polyline):
    """Geometry consisting of connected lines implicitly closed."""

    name = "POLYGON"

    def __init__(self, points: Sequence[Tuple[float, float]]):
        super().__init__(points)

    @classmethod
    def from_list(cls, list: Sequence[float]) -> "Polygon":
        coordinates = cls.list_to_coords(list)
        return cls(coordinates)

    @classmethod
    def from_dict(
        cls,
        dictionary: Union[
            Dict[str, Union[str, float]], Sequence[Dict[str, Union[str, float]]]
        ],
        x: str,
        y: str,
    ) -> "Polygon":
        coordinates = cls._coordinates_from_dict(dictionary, x, y)
        return cls(coordinates)

    def __repr__(self) -> str:
        return f"Polygon({self.to_coords()})"


class Annotation:
    def __init__(
        self, geometry: Geometry, measurements: Optional[Sequence[Measurement]] = None
    ):
        """Represents an annotation, with geometry and an optional list of
        measurements.

        Parameters
        ----------
        geometry: Geometry
            Geometry of the annotation.
        measurements: Optional[Sequence[Measurement]]
            Optional measurements of the annotation.
        """

        self._geometry = geometry
        if measurements is None:
            measurements = []
        self._measurements = measurements

    def __repr__(self) -> str:
        return f"Annotation({self.geometry}, {self.measurements})"

    def __eq__(self, other: "Annotation") -> bool:
        if not isinstance(other, Annotation):
            return NotImplemented
        return (
            self.geometry == other.geometry and self.measurements == other.measurements
        )

    @property
    def geometry(self) -> Geometry:
        return self._geometry

    @property
    def measurements(self) -> Sequence[Measurement]:
        return self._measurements

    def get_measurements(
        self, code: MeasurementCode, unit: UnitCode
    ) -> List[Measurement]:
        """Return measurements of specified type.

        Parameters
        ----------
        code: MeasurementCode
            Code of measurement type to get.
        unit: UnitCode
            Unit of measurement type to get.

        Returns
        -------
        List[Measurement]
            List of measurements of specified type.
        """
        return [
            measurement
            for measurement in self.measurements
            if measurement.same_type(code, unit)
        ]

    def get_measurement_values(
        self, code: MeasurementCode, unit: UnitCode
    ) -> List[float]:
        """Return values for measurements of specified type.

        Parameters
        ----------
        code: MeasurementCode
            Code of measurement type to get.
        unit: UnitCode
            Unit of measurement type to get.

        Returns
        -------
        List[Measurement]
            List of values for measurements of specified type.
        """
        return [
            measurement.value
            for measurement in self.measurements
            if measurement.same_type(code, unit)
        ]


GeometryType = TypeVar("GeometryType", bound=Geometry)
AnnotationGroupType = TypeVar("AnnotationGroupType", bound="AnnotationGroup")


class AnnotationGroup(Generic[GeometryType]):
    _geometry_type: Type[Geometry]

    def __init__(
        self,
        annotations: Sequence[Annotation],
        label: str,
        category_code: AnnotationCategoryCode,
        type_code: AnnotationTypeCode,
        description: Optional[str] = None,
        color: Optional[LabColor] = None,
        is_double: bool = True,
        instance: Optional[UID] = None,
    ):
        """Represents a group of annotations of the same type.

        Parameters
        ----------
        annotations: Sequence[Annotation]
            Annotations in the group.
        label: str
            Group label
        category_code: AnnotationCategoryCode
            Group category code.
        type_code: AnnotationTypeCode
            Group type code.
        description: Optional[str] = None
            Group description.
        color: Optional[LabColor] = None
            Recommended CIELAB color.
        is_double: bool
            If group is stored with double float
        instance: Optional[Uid]
            Uid this group was created from.

        """
        self.validate_type(annotations)
        self._z_planes: List[float] = []
        self._optical_paths: List[str] = []
        self._uid = generate_uid()
        self._category_code = category_code
        self._type_code = type_code
        self._is_double = is_double
        if self._is_double:
            self._point_data_type = np.float64
        else:
            self._point_data_type = np.float32

        self._annotations = annotations
        self._label = label
        self._description = description
        self._color = color
        self._instance = instance

    def __len__(self) -> int:
        return len(self.annotations)

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self.annotations}, {self.label}, "
            f"{self.category_code}, {self.type_code}, "
            f"{self.description}, {self._is_double})"
        )

    def __eq__(self, other: "AnnotationGroup") -> bool:
        if not isinstance(other, AnnotationGroup):
            return NotImplemented
        return (
            self.annotations == other.annotations
            and self.label == other.label
            and self.type_code == other.type_code
            and self.category_code == other.category_code
            and self.description == other.description
            and self.color == other.color
            and self._is_double == other._is_double
        )

    @property
    def category_code(self) -> AnnotationCategoryCode:
        return self._category_code

    @property
    def type_code(self) -> AnnotationTypeCode:
        return self._type_code

    @property
    def annotations(self) -> Sequence[Annotation]:
        return self._annotations

    @property
    def label(self) -> str:
        return self._label

    @property
    def description(self) -> Optional[str]:
        return self._description

    @property
    def color(self) -> Optional[LabColor]:
        return self._color

    @property
    def point_coordinates_data(self) -> np.ndarray:
        """Return coordinates for annotations in group np array.

        Returns
        -------
        np.ndarray
            Coordinates in annotation group.
        """
        coordinate_list: List[float] = []
        for annotation in self.annotations:
            coordinate_list += annotation.geometry.data
        return np.array(coordinate_list, dtype=self._point_data_type)

    @property
    def number_of_annotations(self) -> int:
        """Return number of annotations in group.

        Returns
        -------
        int
            Number of annotations in group.
        """
        return len(self.annotations)

    @property
    def geometry_type(self) -> Type[Geometry]:
        return self._geometry_type

    @property
    def annotation_type(self) -> str:
        return self._geometry_type.name

    def __getitem__(
        self, index: Union[int, Sequence[int]]
    ) -> Union[Annotation, List[Annotation]]:
        if isinstance(index, Sequence):
            return [self.annotations[i] for i in index]
        return self.annotations[index]

    @classmethod
    def from_ds(
        cls: Type[AnnotationGroupType], ds: Dataset, instance: UID
    ) -> AnnotationGroupType:
        """Return annotation group from Annotation Group Sequence dataset.

        Parameters
        ----------
        ds: Dataset
            Dataset containing annotation group.
        instance: Uid
            Uid this group was created from.

        Returns
        -------
        AnnotationGroupType
            Annotation group from dataset.
        """
        is_double = cls._is_ds_double(ds)
        annotations = cls._get_annotations_from_ds(ds)
        label = cls._get_label_from_ds(ds)
        description = getattr(ds, "AnnotationGroupDescription", None)
        type_code = AnnotationTypeCode.from_ds(ds)
        if type_code is None:
            raise ValueError(f"Dataset is missing {AnnotationTypeCode.sequence_name}.")
        category_code = AnnotationCategoryCode.from_ds(ds)
        if category_code is None:
            raise ValueError(
                f"Dataset is missing {AnnotationCategoryCode.sequence_name}."
            )
        color = getattr(ds, "RecommendedDisplayCIELabValue", None)
        return cls(
            annotations,
            label,
            category_code,
            type_code,
            description,
            color,
            is_double,
            instance,
        )

    @staticmethod
    def _get_label_from_ds(ds: Dataset) -> str:
        """Return group label from dataset.

        Parameters
        ----------
        ds: Dataset
            Dataset containing annotation group label.

        Returns
        -------
        str
            Annotation group label from dataset.
        """
        return str(ds.AnnotationGroupLabel)

    @staticmethod
    def _get_focal_planes_from_ds(ds: Dataset) -> List[float]:
        """Return focal planes from dataset. If annotation applies to all focal
        planes returns empty list.

        Parameters
        ----------
        ds: Dataset
            Dataset containing annotation group focal planes.

        Returns
        -------
        List[float]
            Annotation group focal planes from dataset.
        """
        if ds.AnnotationAppliesToAllZPlanes == "YES":
            return []
        try:
            z_planes: List[float] = ds.CommonZCoordinateValue
        except AttributeError:
            raise NotImplementedError(
                "Only 3D annotations with common z coordinate is supported"
            )
        return z_planes

    @staticmethod
    def _get_optical_paths_from_ds(ds: Dataset) -> List[str]:
        """Return optical paths from dataset. If annotation applies to all
        optical paths returns empty list.

        Parameters
        ----------
        ds: Dataset
            Dataset containing annotation group optical paths.

        Returns
        -------
        List[str]
            Annotation group optical paths from dataset.
        """
        if ds.AnnotationAppliesToAllOpticalPaths == "YES":
            return []
        optical_paths: List[str] = ds.ReferencedOpticalPathIdentifier
        return optical_paths

    @staticmethod
    def _get_count_from_ds(ds: Dataset) -> int:
        """Return number of annotations in group.

        Parameters
        ----------
        ds: Dataset
            Dataset containing annotation group.

        Returns
        -------
        int
            Number of annotations in group
        """
        return int(ds.NumberOfAnnotations)

    @classmethod
    def _get_measurements_from_ds(
        cls,
        ds: Dataset,
    ) -> Dict[int, List[Measurement]]:
        """Return measurements grouped by annotation index from ds.

        Parameters
        ----------
        ds: Dataset
            Dataset containing measurement sequence.

        Returns
        -------
        Dict[int, List[Measurement]]:
            Measurements grouped by annotation index.
        """
        if "MeasurementsSequence" not in ds:
            return {}
        annotation_count = cls._get_count_from_ds(ds)
        return Measurement.get_measurements_from_ds(
            ds.MeasurementsSequence, annotation_count
        )

    @staticmethod
    def _is_ds_double(ds: Dataset) -> bool:
        """Return true if group in dataset stores coordinates as double float.

        Parameters
        ----------
        ds: Dataset
            Dataset containing annotation group.

        Returns
        -------
        bool
            True if group uses double float.
        """
        return "DoublePointCoordinatesData" in ds

    @classmethod
    def _get_coordinates_from_ds(cls, ds: Dataset) -> List[Tuple[float, float]]:
        """Return annotation coordinates.

        Parameters
        ----------
        ds: Dataset
            Dataset containing annotation group.

        Returns
        -------
        List[float]
            The coordinates for the annotations in the group.
        """
        if cls._is_ds_double(ds):
            data = dcm_to_list(ds.DoublePointCoordinatesData, "d")
        else:
            data = dcm_to_list(ds.PointCoordinatesData, "f")
        return Geometry.list_to_coords(data)

    @classmethod
    @abstractmethod
    def _get_geometries_from_ds(cls, ds: Dataset) -> List[Geometry]:
        """Abstract method for getting geometries from dataset.

        Parameters
        ----------
        ds: Dataset
            Dataset containing annotation group.

        Returns
        -------
        List[Geometry]
            Geometries in the annotation group.
        """
        raise NotImplementedError()

    @classmethod
    def _get_annotations_from_ds(cls, ds: Dataset) -> List[Annotation]:
        """Return annotation coordinates.

        Parameters
        ----------
        ds: Dataset
            Dataset containing annotation group.

        Returns
        -------
        List[float]
            The coordinates for the annotations in the group.
        """
        measurements = cls._get_measurements_from_ds(ds)
        geometries = cls._get_geometries_from_ds(ds)

        return [
            Annotation(geometry, measurements.get(i))
            for i, geometry in enumerate(geometries)
        ]

    @property
    def measurement_types(self) -> List[Tuple[MeasurementCode, UnitCode]]:
        """Return measurement types in annotation group.

        Returns
        -------
        List[Tuple[MeasurementCode, UnitCode]]
            Measurement code-unit-pairs in annotation group.
        """
        pairs: Set[Tuple[MeasurementCode, UnitCode]] = set()
        for annotation in self.annotations:
            for measurement in annotation.measurements:
                pair = (measurement.code, measurement.unit)
                pairs.add(pair)
        return list(pairs)

    def get_measurements(
        self, code: MeasurementCode, unit: UnitCode
    ) -> List[Measurement]:
        """Return measurements of specified code-unit-pair.

        Parameters
        ----------
        code: MeasurementCode
            Type of measurement type to get.
        unit: UnitCode
            Unit of measurement type to get.

        Returns
        -------
        List[Measurement]
            Measurements of specified type.
        """
        measurements: List[Measurement] = []
        for annotation in self.annotations:
            measurements += annotation.get_measurements(code, unit)
        return measurements

    def create_measurement_indices(
        self, code: MeasurementCode, unit: UnitCode
    ) -> np.ndarray:
        """Return measurement indices for all measurements of specified type.
        Indices are stored starting at index 1.

        Parameters
        ----------
        code: MeasurementCode
            Type of measurement type to create.
        unit: UnitCode
            Unit of measurement type to create.

        Returns
        -------
        np.ndarray
            Measurement indices in annotation group of specified type.
        """
        indices: List[int] = []
        for i, annotation in enumerate(self.annotations):
            indices += [i + 1] * len(annotation.get_measurements(code, unit))
        return np.array(indices, dtype=np.int32)

    def _create_measurement_values(
        self, code: MeasurementCode, unit: UnitCode
    ) -> np.ndarray:
        """Return measurement values for all measurements of specified type.

        Parameters
        ----------
        code: MeasurementCode
            Code of measurement type to create.
        unit: UnitCode
            Unit of measurement type to create.

        Returns
        -------
        np.ndarray
            Measurement values in annotation group of specified type.
        """
        values: List[float] = []
        for annotation in self.annotations:
            values += annotation.get_measurement_values(code, unit)
        return np.array(values, dtype=np.float32)

    def _measurements_is_one_to_one(
        self, code: MeasurementCode, unit: UnitCode
    ) -> bool:
        """Check if all annotations in group have strictly one measurement
        of specified type-unit-pair.

        Parameters
        ----------
        code: MeasurementCode
            Code of measurement type to check.
        unit: UnitCode
            Unit of measurement type to check.

        Returns
        -------
        bool
            True if this measurement code-unit-pair maps one-to-one to
            annotations.
        """
        for annotation in self.annotations:
            if not len(annotation.get_measurements(code, unit)) == 1:
                return False
        return True

    def _create_measurement_value_sequence(
        self, code: MeasurementCode, unit: UnitCode
    ) -> DicomSequence:
        """Return Measurement Value Sequence of measurements of specified
        code and unit.

        Parameters
        ----------
        code: MeasurementCode
            Code of measurement create sequence for.
        unit: UnitCode
            Unit of measurement create sequence for.

        Returns
        -------
        DicomSequence
            A sequence of the measurement values of specified type and unit.
        """
        measurements_ds = Dataset()
        # According to the standard the indices is not needed if each
        # annotation has one and only one measurement.
        if not self._measurements_is_one_to_one(code, unit):
            indices = self.create_measurement_indices(code, unit)
            measurements_ds.AnnotationIndexList = indices
        values = self._create_measurement_values(code, unit)
        measurements_ds.FloatingPointValues = values
        measurement_value_sequence = DicomSequence([measurements_ds])
        return measurement_value_sequence

    def _create_measurement_sequence_item(
        self, code: MeasurementCode, unit: UnitCode
    ) -> Dataset:
        """Return Measurement Sequence item of measurements of specified code
        and unit.

        Parameters
        ----------
        code: MeasurementCode
            Code of measurement to create sequence for.
        unit: UnitCode
            Unit of measurement to create sequence for.

        Returns
        -------
        Dataset
            A measurement sequence item for the specified measurement type and
            unit.
        """
        ds = Dataset()
        ds.MeasurementValuesSequence = self._create_measurement_value_sequence(
            code, unit
        )
        code.insert_into_ds(ds)
        unit.insert_into_ds(ds)
        return ds

    def _set_planes_in_ds(self, ds: Dataset) -> Dataset:
        """Insert the group focal plane attributes into the Annotation Group
        Sequence.

        Parameters
        ----------
        ds: Dataset
            The Annotation Group Sequence.

        Returns
        -------
        Dataset
            The Annotation Group Sequence with focal plane attributes.
        """
        if len(self._z_planes) == 0:
            ds.AnnotationAppliesToAllZPlanes = "YES"
        else:
            ds.AnnotationAppliesToAllZPlanes = "NO"
            ds.CommonZCoordinateValue = self._z_planes
        return ds

    def _set_optical_paths_in_ds(self, ds: Dataset) -> Dataset:
        """Insert the group optical path attributes into the Annotation Group
        Sequence.

        Parameters
        ----------
        ds: Dataset
            The Annotation Group Sequence.

        Returns
        -------
        Dataset
            The Annotation Group Sequence with optical path attributes.
        """
        if len(self._optical_paths) == 0:
            ds.AnnotationAppliesToAllOpticalPaths = "YES"
        else:
            ds.AnnotationAppliesToAllOpticalPaths = "NO"
            ds.ReferencedOpticalPathIdentifier = self._optical_paths
        return ds

    def _set_coordinates_data_in_ds(self, ds: Dataset) -> Dataset:
        """Insert the group point coordinates into the Annotation Group
        Sequence.

        Parameters
        ----------
        ds: Dataset
            The Annotation Group Sequence.

        Returns
        -------
        Dataset
            The Annotation Group Sequence with inserted coordinates.
        """
        if self._is_double:
            ds.DoublePointCoordinatesData = self.point_coordinates_data
        else:
            ds.PointCoordinatesData = self.point_coordinates_data
        return ds

    def _set_measurement_sequence_in_ds(self, ds: Dataset) -> Dataset:
        """Insert group measurements into the Annotation Group Sequence.

        Parameters
        ----------
        ds: Dataset
            The Annotation Group Sequence.

        Returns
        -------
        Dataset
            The Annotation Group Sequence with inserted measurements.
        """
        ds.MeasurementsSequence = DicomSequence(
            [
                self._create_measurement_sequence_item(*measurement_type)
                for measurement_type in self.measurement_types
            ]
        )
        return ds

    def to_ds(self, group_number: int) -> Dataset:
        """Return annotation group as a Annotation Group Sequence
        item.

        Parameters
        ----------
        group_number: int
            The index of the group.

        Returns
        -------
        Dataset
            Dataset containing the group.
        """
        ds = Dataset()
        ds.AnnotationGroupNumber = group_number
        ds.AnnotationGroupUID = self._uid
        ds.NumberOfAnnotations = len(self.annotations)
        ds.GraphicType = self.annotation_type
        ds.AnnotationGroupLabel = self.label
        if self.description is not None:
            ds.AnnotationGroupDescription = self.description
        ds = self.category_code.insert_into_ds(ds)
        ds = self.type_code.insert_into_ds(ds)
        ds = self._set_coordinates_data_in_ds(ds)
        ds = self._set_planes_in_ds(ds)
        ds = self._set_optical_paths_in_ds(ds)
        ds = self._set_measurement_sequence_in_ds(ds)
        if self.color is not None:
            ds.RecommendedDisplayCIELabValue = self.color
        # AUTOMATIC and SEMIAUTOMATIC requires a
        # Annotation Algorithm Identification Sequence
        ds.AnnotationGroupGenerationType = "MANUAL"
        return ds

    @classmethod
    def validate_type(cls, annotations: Sequence[Annotation]):
        """Check that list of annotations are of the requested type.

        Parameters
        ----------
        annotations: Sequence[Annotation]
            List of annotations to check
        """
        for annotation in annotations:
            if not isinstance(annotation.geometry, cls._geometry_type):
                raise TypeError(
                    f"annotation type {type(annotation.geometry)}"
                    f" does not match Group type code {cls._geometry_type}"
                )

    @classmethod
    def _get_group_type_by_geometry(
        cls, geometry_type: Type[Geometry]
    ) -> Type["AnnotationGroup"]:
        """Return AnnotationGroup class for geometry type.

        Parameters
        ----------
        geometry_type: type
            The geometry type to get AnnotationGroup class for.

        """
        if geometry_type == Point:
            return PointAnnotationGroup
        elif geometry_type == Polyline:
            return PolylineAnnotationGroup
        elif geometry_type == Polygon:
            return PolygonAnnotationGroup
        raise NotImplementedError(f"{type(geometry_type)} is not supported")

    @classmethod
    def from_geometries(
        cls,
        geometries: Sequence[Geometry],
        label: str,
        category_code: AnnotationCategoryCode,
        type_code: AnnotationTypeCode,
        is_double: bool = True,
    ) -> "AnnotationGroup":
        """Return AnnotationGroup created from list of geometries. The group
        type is determined by the first geometry, and all geometries needs to
        have the same type.

        Parameters
        ----------
        geometries: Sequence[Geometries]
            Geometries in the group.
        label: str
            Group label
        category_code: AnnotationCategoryCode
            Group category code.
        type_code: AnnotationTypeCode
            Group type code.

        Returns
        -------
        AnnotationGroup
            Group created from geometries.

        """
        geometry_type = type(geometries[0])
        group_type = cls._get_group_type_by_geometry(geometry_type)
        group: AnnotationGroup = group_type(
            annotations=[Annotation(geometry) for geometry in geometries],
            label=label,
            category_code=category_code,
            type_code=type_code,
            is_double=is_double,
        )
        return group


class PointAnnotationGroup(AnnotationGroup[Point]):
    """Point annotation group"""

    _geometry_type = Point

    @classmethod
    def _get_geometries_from_ds(cls, ds: Dataset) -> List[Point]:
        """Returns point geometries from dataset.

        Parameters
        ----------
        ds: Dataset
            Dataset containing annotation group.

        Returns
        -------
        List[Geometry]
            Point geometries in the annotation group.
        """
        coordinate_list = cls._get_coordinates_from_ds(ds)
        points = [Point(*coordinates) for coordinates in coordinate_list]
        return points


class PolylineAnnotationGroupMeta(AnnotationGroup[GeometryType]):
    """Meta class for line annotation group"""

    geometry_type: Type[Geometry]

    @property
    def point_index_list(self) -> np.ndarray:
        """Return point index list for annotations in group. Indices are stored
        starting at index 1 and in relation to geometry data length.

        Returns
        -------
        np.ndarray
            List of indices in annotation group
        """
        index = 1
        indices: List[int] = []
        for annotation in self.annotations:
            indices.append(index)
            index += len(annotation.geometry.data)
        return np.array(indices, dtype=np.int32)

    @staticmethod
    def _get_indices_from_ds(ds: Dataset) -> List[int]:
        """Return line start indices from sup 222 dataset. Indices are stored
        starting at with value 1, and are in relation to non-pared coordinates.
        Returned list starts at 0 and is in relation to paired coordinates.

        Parameters
        ----------
        ds: Dataset
            Dataset containing sup 222 indices.

        Returns
        -------
        List[int]
            List of indices in dataset.
        """
        return [
            (value - 1) // 2
            for value in dcm_to_list(ds.LongPrimitivePointIndexList, "l")
        ]

    @classmethod
    def _get_geometries_from_ds(cls, ds: Dataset) -> List[GeometryType]:
        """Returns line geometries from dataset. Each line geometry consists of
        multiple points, and the first coordinate in the coordinate list is

        Parameters
        ----------
        ds: Dataset
            Dataset containing annotation group.

        Returns
        -------
        List[Geometry]
            Polyline geometries in the annotation group.
        """
        indices = cls._get_indices_from_ds(ds)
        number_of_geometries = ds.NumberOfAnnotations
        if number_of_geometries != len(indices):
            raise ValueError(
                "Number of indices must be the same as number of geometries"
            )
        coordinates = cls._get_coordinates_from_ds(ds)
        indices += [len(coordinates)]  # Add end for last geometry
        geometries: List[GeometryType] = []
        for index in range(number_of_geometries):
            start = indices[index]
            end = indices[index + 1]
            line_coordinates = coordinates[start:end]
            lines = cls._get_line_geometry_from_coords(line_coordinates)
            geometries.append(lines)
        return geometries

    @staticmethod
    @abstractmethod
    def _get_line_geometry_from_coords(
        coordinates: Sequence[Tuple[float, float]]
    ) -> GeometryType:
        raise NotImplementedError()

    def to_ds(self, group_number: int) -> Dataset:
        """Return annotation group as a Annotation Group Sequence item.

        Parameters
        ----------
        group_number: int
            The index of the group.

        Returns
        -------
        Dataset
            Dataset containing the group.
        """
        ds = super().to_ds(group_number)
        ds.LongPrimitivePointIndexList = self.point_index_list
        return ds


class PolylineAnnotationGroup(PolylineAnnotationGroupMeta[Polyline]):
    _geometry_type = Polyline

    @staticmethod
    def _get_line_geometry_from_coords(coordinates: Sequence[Tuple[float, float]]):
        return Polyline(coordinates)


class PolygonAnnotationGroup(PolylineAnnotationGroupMeta[Polygon]):
    _geometry_type = Polygon

    @staticmethod
    def _get_line_geometry_from_coords(coordinates: Sequence[Tuple[float, float]]):
        return Polygon(coordinates)


class AnnotationInstance:
    """Class for handling microscope bulk simple annotations according to
    sup-222. Point, polyline, and polygon graphic types are implemented,
    ellipse and rectangle graphic types are not implemented. Annotation must
    have common z-coordinate or apply to all z planes (annotation with 3D
    PointCoordinateData is not implemented.)
    """

    def __init__(
        self,
        groups: Sequence[AnnotationGroup],
        coordinate_type: str,
        slide_uids: SlideUids,
    ):
        """Represents a collection of annotation groups.

        Parameters
        ----------
        annotations: Sequence[AnnotationGroup]
            List of annotations group
        coordinate_type: str
            If coordinates are volume-related ('volume') or image-related
            ('image').
        frame_of_referenc: Uid
            Frame of reference uid of image that the annotations belong to
        """
        self.groups = groups
        if coordinate_type not in ["image", "volume"]:
            raise ValueError("Coordinate type should be 'image' or 'volume'")
        self.coordinate_type = coordinate_type
        self.slide_uids = slide_uids
        self.datetime = datetime.now()
        self.modality = "ANN"
        self.series_number: int

    def __repr__(self) -> str:
        return f"AnnotationInstance({self.groups}, " f"{self.slide_uids})"

    def save(
        self,
        path: Union[str, Path],
        little_endian: bool = True,
        implicit_vr: bool = False,
        uid_generator: Callable[..., UID] = generate_uid,
    ):
        """Write annotations to DICOM file according to sup 222.
        Note that the file will miss DICOM attributes that has not yet been
        implemented.

        Parameters
        ----------
        path: Union[str, Path]
            Path to write DICOM file to
        little_endian: bool
            Write DICOM file as little endian
        implicit_vr: bool
            Write DICOM file with implicit value representation
        """
        ds = Dataset()
        ds.is_little_endian = little_endian
        ds.is_implicit_VR = implicit_vr
        bulk_sequence = DicomSequence()
        for index, annotation_group in enumerate(self.groups):
            if isinstance(annotation_group, AnnotationGroup):
                bulk_sequence.append(annotation_group.to_ds(index + 1))
            else:
                raise NotImplementedError(
                    f"Group type: {type(annotation_group)} not supported"
                )
        ds.AnnotationGroupSequence = bulk_sequence
        if self.coordinate_type == "image":
            ds.AnnotationCoordinateType = "2D"
        elif self.coordinate_type == "volume":
            ds.AnnotationCoordinateType = "3D"
        if self.slide_uids.frame_of_reference is not None:
            ds.FrameOfReferenceUID = self.slide_uids.frame_of_reference
        ds.StudyInstanceUID = self.slide_uids.study_instance
        ds.SeriesInstanceUID = self.slide_uids.series_instance
        ds.SOPInstanceUID = uid_generator()
        ds.SOPClassUID = ANN_SOP_CLASS_UID

        meta_ds = FileMetaDataset()
        if little_endian and implicit_vr:
            transfer_syntax = ImplicitVRLittleEndian
        elif little_endian and not implicit_vr:
            transfer_syntax = ExplicitVRLittleEndian
        elif not little_endian and not implicit_vr:
            transfer_syntax = ExplicitVRBigEndian
        else:
            raise NotImplementedError("Unsupported transfer syntax")

        meta_ds.TransferSyntaxUID = transfer_syntax
        meta_ds.MediaStorageSOPInstanceUID = ds.SOPInstanceUID
        meta_ds.MediaStorageSOPClassUID = ANN_SOP_CLASS_UID
        meta_ds.FileMetaInformationGroupLength = 0  # Updated on write
        validate_file_meta(meta_ds)
        file_ds = FileDataset(
            preamble=b"\x00" * 128,
            filename_or_obj=path,
            file_meta=meta_ds,
            dataset=ds,
            is_implicit_VR=implicit_vr,
            is_little_endian=little_endian,
        )
        dcmwrite(path, file_ds)

    @classmethod
    def open(
        cls, files: Iterable[Union[str, Path, BinaryIO]]
    ) -> Iterable["AnnotationInstance"]:
        """Read annotations from DICOM files according to sup 222.

        Parameters
        ----------
        files: Sequence[Union[str, Path, BinaryIO]]
            Files with DICOM annotations to read.
        """
        return (cls.open_dataset(dcmread(file)) for file in files)

    @classmethod
    def open_dataset(cls, dataset: Dataset) -> "AnnotationInstance":
        """Read annotations from DICOM dataset according to sup 222.

        Parameters
        ----------
        dataset: Dataset
            DICOM annotation dataset to read.

        Returns
        -------
        List[AnnotationGroup]
            Annotation groups read from dataset.
        """
        groups: List[AnnotationGroup] = []
        slide_uids: Optional[SlideUids] = None

        frame_of_reference_uid = getattr(dataset, "FrameOfReferenceUID", None)
        if dataset.AnnotationCoordinateType == "2D":
            coordinate_type = "image"
        elif dataset.AnnotationCoordinateType == "3D":
            coordinate_type = "volume"
        else:
            raise ValueError("Unknown coordinate type")
        if coordinate_type == "volume" and frame_of_reference_uid is None:
            raise ValueError(
                "volume annotation corrindate type requires frame of reference"
            )

        instance = dataset.SOPInstanceUID
        if slide_uids is None:
            slide_uids = SlideUids(
                dataset.StudyInstanceUID,
                dataset.SeriesInstanceUID,
                frame_of_reference_uid,
            )
        else:
            if slide_uids != SlideUids(
                dataset.StudyInstanceUID,
                dataset.SeriesInstanceUID,
                frame_of_reference_uid,
            ):
                raise ValueError("Base uids should match")
        for annotation_ds in dataset.AnnotationGroupSequence:
            annotation_type = annotation_ds.GraphicType
            if annotation_type == "POINT":
                annotation_class = PointAnnotationGroup
            elif annotation_type == "POLYLINE":
                annotation_class = PolylineAnnotationGroup
            elif annotation_type == "POLYGON":
                annotation_class = PolygonAnnotationGroup
            else:
                raise NotImplementedError("Unsupported Graphic type")
            annotation = annotation_class.from_ds(annotation_ds, instance)
            groups.append(annotation)
        return cls(groups, coordinate_type, slide_uids)

    def __getitem__(self, index: int) -> AnnotationGroup:
        """Return annotation group by index (group number).

        Parameters
        ----------
        index: int
            Group number of group to return

        Returns
        -------
        Annotation
            Annotation group with group number
        """
        return self.groups[index]

    def __len__(self) -> int:
        return len(self.groups)



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from collections import UserDict
from typing import (
    Callable,
    Iterable,
    List,
    Union,
)

from wsidicom.metadata.schema.json.sample.model import (
    BaseSpecimenJsonModel,
)
from wsidicom.metadata.sample import (
    SampledSpecimen,
    BaseSpecimen,
    SpecimenIdentifier,
)


class SpecimenDictionary(UserDict[Union[str, SpecimenIdentifier], BaseSpecimen]):
    """Dictionary for specimens that creates missing specimens when accessed."""

    def __init__(
        self,
        create_missing: Callable[[Union[str, SpecimenIdentifier]], BaseSpecimen],
        *args,
        **kwargs,
    ):
        """Initialize dictionary.

        Parameters
        ----------
        create_missing: Callable[[Union[str, SpecimenIdentifier]], Specimen]
            Function that creates a specimen for a given identifier.
        """
        super().__init__(*args, **kwargs)
        self._create_missing = create_missing

    def __missing__(self, key: Union[str, SpecimenIdentifier]) -> BaseSpecimen:
        """Create missing specimen."""
        specimen = self._create_missing(key)
        self[key] = specimen
        return specimen


class SpecimenJsonParser:
    """Factory for creating specimens from json models."""

    def __init__(self, specimen_models: Iterable[BaseSpecimenJsonModel]):
        """Initiate factory.

        Parameters
        ----------
        specimen_models: Iterable[BaseSpecimenJsonModel]
            Json models to create specimens from.
        """
        self._specimens = SpecimenDictionary(self._make_specimen)
        self._specimen_models_by_identifier = {
            specimen.identifier: specimen for specimen in specimen_models
        }

    def create_specimens(self) -> List[BaseSpecimen]:
        """Create specimens

        Returns
        -------
        List[Specimen]
            List of created specimens.
        """
        for identifier in self._specimen_models_by_identifier:
            self._specimens[identifier] = self._make_specimen(identifier)
        sampled_specimens = [
            sampled_from.specimen.identifier
            for specimen in self._specimens.values()
            if isinstance(specimen, SampledSpecimen)
            for sampled_from in specimen.sampled_from_list
        ]
        return [
            specimen
            for specimen in self._specimens.values()
            if specimen.identifier not in sampled_specimens
        ]

    def _make_specimen(
        self,
        identifier: Union[str, SpecimenIdentifier],
    ) -> BaseSpecimen:
        """Create specimen by identifier from json model.

        Create nested specimens that the specimen is sampled from if needed.

        Parameters
        ----------
        identifier: Union[str, SpecimenIdentifier]
            Identifier of specimen to create

        Returns
        -------
        Specimen
            Specimen created from json model.

        """
        specimen_model = self._specimen_models_by_identifier[identifier]
        return specimen_model.from_json_model(self._specimens)



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Index for frame positions and length in image data."""

from abc import abstractmethod
from functools import cached_property
from typing import List, Optional, Tuple

from wsidicom.errors import WsiDicomFileError
from wsidicom.file.io.frame_index.offset_table_type import OffsetTableType
from wsidicom.file.io.wsidicom_io import WsiDicomIO
from wsidicom.tags import PixelDataTag


class FrameIndex:
    def __init__(self, file: WsiDicomIO, pixel_data_start: int, frame_count: int):
        self._file = file
        self._frame_count = frame_count
        self._pixel_data_start = pixel_data_start
        self._file.seek(self._pixel_data_start)
        self._pixels_start = self._get_pixels_start()

    @cached_property
    def index(self) -> List[Tuple[int, int]]:
        """Return a list of frame positions and lengths."""
        self._file.seek(self._pixel_data_start)
        index = self._get_index()
        self._validate_frame_index(index)
        return index

    @property
    @abstractmethod
    def offset_table_type(self) -> OffsetTableType:
        raise NotImplementedError()

    @abstractmethod
    def _get_index(self) -> List[Tuple[int, int]]:
        """Return a list of frame positions and lengths."""
        raise NotImplementedError()

    @abstractmethod
    def _get_pixels_start(self) -> int:
        """Parse pixel data start and return position of first frame."""
        raise NotImplementedError()

    def _validate_frame_index(self, frame_index: List[Tuple[int, int]]):
        """Validate frame index.

        Parameters
        ----------
        frame_index: List[Tuple[int, int]]
            Frame index.
        """
        if len(frame_index) < self._frame_count:
            raise WsiDicomFileError(
                str(self._file),
                (
                    f"ImageData contained less frames {len(frame_index)} than "
                    f"NumberOfFrames {self._frame_count}."
                ),
            )
        if len(frame_index) > self._frame_count:
            raise WsiDicomFileError(
                str(self._file),
                (
                    f"ImageData contained more fragments {len(frame_index)} than "
                    f"NumberOfFrames {self._frame_count} and fragmented frames are not "
                    "supported."
                ),
            )

    def _validate_pixel_data_start(self, expected_length: Optional[int]):
        """Check that pixel data tag is present and that the tag length is equal to
        expected count. Raises WsiDicomFileError otherwise.

        """
        tag = self._file.read_tag()
        if tag != PixelDataTag:
            WsiDicomFileError(str(self._file), "Expected PixelData tag")
        self._file.read_tag_vr()
        length = self._file.read_tag_length(True)
        if expected_length is not None and length != expected_length:
            raise WsiDicomFileError(
                str(self._file),
                f"Expected {expected_length} length when reading Pixel data, got {length}.",
            )
        elif expected_length is None and length != 0xFFFFFFFF:
            raise WsiDicomFileError(
                str(self._file),
                f"Expected undefined length when reading Pixel data, got {length}.",
            )



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module with DICOM specimen description parser."""

import logging
from dataclasses import dataclass
from typing import (
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from wsidicom.conceptcode import (
    AnatomicPathologySpecimenTypesCode,
    ContainerTypeCode,
)
from wsidicom.config import settings
from wsidicom.metadata.sample import (
    BaseSampling,
    BaseSpecimen,
    Collection,
    Embedding,
    Fixation,
    PreparationStep,
    Processing,
    Receiving,
    Sample,
    SlideSample,
    Specimen,
    SpecimenIdentifier,
    Staining,
    Storage,
)
from wsidicom.metadata.schema.dicom.sample.model import (
    CollectionDicomModel,
    ProcessingDicomModel,
    ReceivingDicomModel,
    SamplingDicomModel,
    SpecimenDescriptionDicomModel,
    SpecimenPreparationStepDicomModel,
    StainingDicomModel,
    StorageDicomModel,
)


@dataclass(frozen=True)
class ParsedSpecimen:
    """Parsed specimen steps."""

    preparation_steps: List[PreparationStep]
    sampling: Optional[BaseSampling]
    container: Optional[ContainerTypeCode]
    specimen_type: Optional[AnatomicPathologySpecimenTypesCode]


class StepsDirectory:
    """Class for holding specimen preparation steps ordered by specimen identifier."""

    def __init__(self, steps: List[SpecimenPreparationStepDicomModel]):
        """Initialize StepsDirectory.

        Parameters
        ----------
        steps: List[SpecimenPreparationStepDicomModel]
            Steps to order.
        """
        self._steps_by_identifier = self._order_steps_by_identifier(steps)

    def get(
        self, identifier: SpecimenIdentifier
    ) -> List[SpecimenPreparationStepDicomModel]:
        """Get steps matching identifier.

        Parameters
        ----------
        identifier: SpecimenIdentifier
            Specimen identifier to get steps for.

        Returns
        -------
        List[SpecimenPreparationStepDicomModel]
            Steps for specimen.
        """
        return self._get_and_clear_steps_for_identifier(identifier)

    def get_previous_specimen_identifier(
        self, identifier: SpecimenIdentifier
    ) -> Optional[SpecimenIdentifier]:
        """Get identifier of previous specimen.

        Parameters
        ----------
        identifier: SpecimenIdentifier
            Identifier of specimen to get previous specimen for.

        Returns
        -------
        Optional[SpecimenIdentifier]
            Identifier of previous specimen, if any.
        """
        matching_identifier = self._get_identifier(
            identifier, self._steps_by_identifier
        )
        if matching_identifier is None:
            return None
        this_specimen_index = list(self._steps_by_identifier.keys()).index(
            matching_identifier
        )
        if this_specimen_index == 0:
            return None
        return list(self._steps_by_identifier.keys())[this_specimen_index - 1]

    @property
    def is_empty(self) -> bool:
        return all(steps is None for steps in self._steps_by_identifier.values())

    def _order_steps_by_identifier(
        self,
        steps: List[SpecimenPreparationStepDicomModel],
    ) -> Dict[SpecimenIdentifier, Optional[List[SpecimenPreparationStepDicomModel]]]:
        """
        Parse steps into dictionary of steps by specimen identifier.

        Parameters
        ----------
        steps: List[SpecimenPreparationStepDicomModel]
            Steps to parse.

        Returns
        -------
        Dict[SpecimenIdentifier, Optional[List[SpecimenPreparationStepDicomModel]]]:
            Steps ordered by specimen identifier.
        """
        steps_by_identifier: Dict[
            SpecimenIdentifier, Optional[List[SpecimenPreparationStepDicomModel]]
        ] = {}
        for step in steps:
            identifier = step.specimen_identifier
            matching_identifier = self._get_identifier(identifier, steps_by_identifier)
            if matching_identifier not in steps_by_identifier:
                steps_by_identifier[identifier] = [step]
            else:
                existing_steps = steps_by_identifier[matching_identifier]
                if existing_steps is None:
                    steps_by_identifier[matching_identifier] = [step]
                else:
                    existing_steps.append(step)
        return steps_by_identifier

    def _get_and_clear_steps_for_identifier(
        self,
        identifier: SpecimenIdentifier,
    ) -> List[SpecimenPreparationStepDicomModel]:
        """Return and clear steps matching identifier.

        If strict matching return steps by key. Otherwise return steps with relaxed
        matching.

        Steps are cleared after being returned to ensure that they are not parsed
        multiple times. The key remains in the dictionary to enable checking that all
        identifiers have been parsed.

        Parameters
        ----------
        identifier: SpecimenIdentifier
            Identifier to match.

        Returns
        -------
        List[SpecimenPreparationStepDicomModel]
            Steps matching identifier.
        """
        matching_identifier = self._get_identifier(
            identifier, self._steps_by_identifier
        )
        if matching_identifier not in self._steps_by_identifier:
            # Specimen does not have any describing preparation steps
            return []
        steps = self._steps_by_identifier[matching_identifier]
        if steps is None:
            raise ValueError(
                f"Tried to get steps for identifier {matching_identifier} that "
                "has already been parsed."
            )
        # Clear steps for identifier, as the specimen should not be parsed again.
        self._steps_by_identifier[matching_identifier] = None
        return steps

    @staticmethod
    def _get_identifier(
        identifier: SpecimenIdentifier,
        steps_by_identifier: Dict[
            SpecimenIdentifier, Optional[List[SpecimenPreparationStepDicomModel]]
        ],
    ) -> Optional[SpecimenIdentifier]:
        """Get identifier matching identifier.

        Parameters
        ----------
        identifier: SpecimenIdentifier
            Identifier to match.
        steps_by_identifier: Dict[
            SpecimenIdentifier,
            Optional[List[SpecimenPreparationStepDicomModel]]
        ]
            Steps ordered by identifier.

        Returns
        -------
        Optional[SpecimenIdentifier]
            Identifier matching identifier.
        """

        if (
            settings.strict_specimen_identifier_check
            or identifier in steps_by_identifier
        ):
            return identifier
        return next(
            (
                existing_identifier
                for existing_identifier in steps_by_identifier.keys()
                if existing_identifier.relaxed_matching(identifier)
            ),
            None,
        )


class SpecimenDicomParser:
    """Parser for DICOM Specimen Description models deserialized from DICOM dataset.

    The DICOM Specimen Description contains a list of Specimen Preparation Steps
    describing the preparation of a specimen. The steps are ordered chronologically and
    describes specimen collection, sampling, staining, etc. for (potentially) the whole
    specimen branch from the extracted specimen to the slide sample.

    We assume (and check) that one Specimen Description only describes one branch of the
    specimen tree, i.e. if for example a slide sample is sampled from a block that has
    been prepared from two different extracted specimens, the Specimen Description will
    only contain steps for one of the extracted specimens. There might be another slide
    sample on the slide with a Specimen Description containing steps for the other
    extracted specimen.

    We also assume (and check) that when the sample specimen occurs in multiple
    Specimen Descriptions it has the same steps, i.e. the specimen has not been further
    processed after performing one of the samplings.

    We assume (and check) that all slide samples on a slide have been stained with the
    same stainings.

    When parsing multiple Specimen Descriptions, we merge any specimens that
    have the same identifier, i.e. two slide samples from the same block will have the
    samme block object as parent specimen.

    Any Specimen Preparation Step can specify a processing method, fixative, and/or
    embedding, e.g. a sampling step can optionally also define a fixative. We create
    separate steps for each of these (processing method, fixative, embedding).

    The parsing of the Specimen Description starts with parsing the steps for the
    slide sample, and then recursively parsing the steps for the parent specimen. Parent
    specimens are either identified by a sampling step or by assuming that the specimen
    was sampled from the previous specimen in the Specimen Description (if any). The
    later sampling will produce an `UnknownSampling`, as the sampling step is missing
    properties as `method` and it can't be certain that the found parent specimen is the
    immediate parent specimen or a specimen further up the specimen branch.
    """

    def __init__(self):
        self._parsed_specimens: Dict[SpecimenIdentifier, BaseSpecimen] = {}
        self._parsed_stainings: List[Staining] = []

    def parse_descriptions(
        self, specimen_descriptions: Iterable[SpecimenDescriptionDicomModel]
    ) -> Tuple[List[SlideSample], List[Staining]]:
        """
        Parse specimen descriptions into samples and stainings.

        Parameters
        ----------
        specimen_descriptions: Iterable["SpecimenDescriptionDicomModel"]
            Specimen descriptions to parse.

        Returns
        ----------
        Tuple[List[SlideSample], List[Staining]]
            Samples and stainings parsed from descriptions.

        """
        slide_samples = [
            self._parsed_description(specimen_description)
            for specimen_description in specimen_descriptions
        ]
        return slide_samples, self._parsed_stainings

    def _parsed_description(
        self, description: SpecimenDescriptionDicomModel
    ) -> SlideSample:
        """
        Create a SlideSample from Specimen Description.

        Contained parent specimens and stainings are created or updated.

        Parameters
        ----------
        description: SpecimenDescriptionDicomModel,
            Specimen description to parse.

        Returns
        ----------
        SlideSample
            Parsed sample.

        """
        stainings = self._parse_stainings(description.steps)
        if len(self._parsed_stainings) == 0:
            self._parsed_stainings = list(stainings)
        elif any(staining not in self._parsed_stainings for staining in stainings):
            raise NotImplementedError(
                "Handling slide samples with different stainings is not implemented."
            )
        steps_directory = StepsDirectory(description.steps)
        identifier = description.specimen_identifier
        parsed_specimen = self._parse_preparation_steps_for_specimen(
            identifier, steps_directory
        )
        if not steps_directory.is_empty:
            logging.warning(
                "Specimen description contained steps for specimens that were not "
                "parsed. If this is due to steps missing issuer of identifiers, try "
                "setting `settings.strict_specimen_identifier_check` to `False`."
            )
        return SlideSample(
            identifier=identifier.simplify(),
            anatomical_sites=description.anatomical_sites,
            sampled_from=parsed_specimen.sampling,
            uid=description.uid,
            localization=description.localization,
            steps=parsed_specimen.preparation_steps,
        )

    def _parse_stainings(
        self,
        steps: List[SpecimenPreparationStepDicomModel],
    ) -> Iterator[Staining]:
        """
        Parse stainings from steps.

        Parameters
        ----------
        steps: List[SpecimenPreparationStepDicomModel]
            Steps to parse.

        Returns
        ----------
        Iterator[Staining]
            Parsed stainings.

        """
        return (
            Staining(
                step.substances, date_time=step.date_time, description=step.description
            )
            for step in steps
            if isinstance(step, StainingDicomModel)
        )

    def _parse_preparation_steps_for_specimen(
        self, identifier: SpecimenIdentifier, steps_directory: StepsDirectory
    ) -> ParsedSpecimen:
        """
        Parse preparation steps and samplings for a specimen.

        Creates or updates parent specimens.

        Parameters
        ----------
        identifier: SpecimenIdentifier
            The identifier of the specimen to parse.
        steps_directory: StepsDirectory
            DICOM preparation steps ordered by specimen identifier.

        Returns
        ----------
        ParsedSpecimen
            Parsed steps for specimen.

        """
        container: Optional[ContainerTypeCode] = None
        specimen_type: Optional[AnatomicPathologySpecimenTypesCode] = None
        sampling: Optional[BaseSampling] = None
        preparation_steps: List[PreparationStep] = []
        steps = steps_directory.get(identifier)
        for index, step in enumerate(steps):
            parsed_step = self._parse_step(
                index, step, identifier, steps_directory, steps
            )
            preparation_steps.extend(parsed_step.preparation_steps)
            if parsed_step.sampling is not None:
                if sampling is not None:
                    raise ValueError(
                        "Should not be more than one sampling for a specimen."
                    )
                sampling = parsed_step.sampling

            # Update container and specimen type
            if step.specimen_type is not None:
                if specimen_type is not None and specimen_type != step.specimen_type:
                    raise ValueError(
                        f"Got mismatching specimen types {specimen_type} and "
                        f"{step.specimen_type} in steps for specimen {identifier}."
                    )
                specimen_type = step.specimen_type
            if step.container is not None:
                if container is not None and container != step.container:
                    raise ValueError(
                        f"Got mismatching container types {container} and "
                        f"{step.container} in steps for specimen {identifier}."
                    )
                container = step.container
        return ParsedSpecimen(
            preparation_steps=preparation_steps,
            sampling=sampling,
            container=container,
            specimen_type=specimen_type,
        )

    def _parse_step(
        self,
        index: int,
        step: SpecimenPreparationStepDicomModel,
        identifier: SpecimenIdentifier,
        steps_directory: StepsDirectory,
        steps: List[SpecimenPreparationStepDicomModel],
    ) -> ParsedSpecimen:
        """Parse step and return nested preparation steps and samplings.

        Parameters
        ----------
        index: int
            Index of step in list of steps for this identifier.
        step: SpecimenPreparationStepDicomModel
            Step to parse.
        identifier: SpecimenIdentifier
            Identifier of specimen to parse.
        steps_directory: StepsDirectory
            DICOM preparation steps ordered by specimen identifier.
        steps: List[SpecimenPreparationStepDicomModel]
            All steps for this identifier.

        Returns
        -------
        ParsedSpecimen
            Parsed step for specimen.
        """
        preparation_steps: List[PreparationStep] = []
        sampling: Optional[BaseSampling] = None
        # First handle collection and sampling steps that should be on index 0
        if index == 0:
            if isinstance(step, CollectionDicomModel):
                preparation_steps.append(
                    self._create_collection_step(step, identifier, steps)
                )
            elif isinstance(step, SamplingDicomModel):
                sampling = self._create_sampling_step(step, steps_directory)
            else:
                # First step was not a collection step or a sampling step, this specimen
                # thus has no defined parent. Add a unknown sampling to previous
                # specimen if any
                sampling = self._create_unknown_sampling(identifier, steps_directory)

        if isinstance(step, ProcessingDicomModel):
            if step.processing is not None:
                # Only add processing step if has processing code.
                preparation_steps.append(
                    Processing(step.processing, step.date_time, step.description)
                )
        elif isinstance(step, ReceivingDicomModel):
            preparation_steps.append(Receiving(step.date_time, step.description))
        elif isinstance(step, StorageDicomModel):
            preparation_steps.append(Storage(step.date_time, step.description))
        elif isinstance(step, StainingDicomModel):
            # Stainings are handled elsewhere
            pass
        elif isinstance(step, (CollectionDicomModel, SamplingDicomModel)):
            if index != 0:
                raise ValueError(
                    f"Found unexpected step of type {type(step)} at index {index}. "
                    "Collection and sampling steps should only be on index 0"
                )
        else:
            raise NotImplementedError(f"Step of type {type(step)} at index {index}.")
        # Fixative, embedding, and processing (for other than processing step)
        # are parsed into separate steps
        if step.fixative is not None:
            preparation_steps.append(
                Fixation(
                    fixative=step.fixative,
                    date_time=step.date_time,
                    description=step.description,
                )
            )
        if step.embedding is not None:
            preparation_steps.append(
                Embedding(
                    medium=step.embedding,
                    date_time=step.date_time,
                    description=step.description,
                )
            )
        if step.processing is not None and not isinstance(step, ProcessingDicomModel):
            preparation_steps.append(
                Processing(
                    step.processing,
                    date_time=step.date_time,
                    description=step.description,
                )
            )
        return ParsedSpecimen(
            preparation_steps=preparation_steps,
            sampling=sampling,
            container=step.container,
            specimen_type=step.specimen_type,
        )

    def _create_collection_step(
        self,
        step: CollectionDicomModel,
        identifier: SpecimenIdentifier,
        steps: List[SpecimenPreparationStepDicomModel],
    ) -> Collection:
        """Create collection step from collection model.

        Parameters
        ----------
        step: CollectionDicomModel
            Collection model to create step from.
        identifier: SpecimenIdentifier
            Identifier of the specimen the collection belongs to.
        steps_directory: StepsDirectory
            DICOM preparation steps ordered by specimen identifier.
        steps: List[SpecimenPreparationStepDicomModel]
            All steps for this identifier.

        Returns
        -------
        Collection
            Collection step created from model.
        """
        any_sampling_steps = any(
            sampling_step
            for sampling_step in steps
            if sampling_step is not None
            and isinstance(sampling_step, SamplingDicomModel)
            and sampling_step.identifier == identifier
        )
        if any_sampling_steps:
            raise ValueError(
                (
                    "Collection step should be first step and there should not "
                    "be any sampling steps."
                )
            )
        return Collection(
            step.method,
            date_time=step.date_time,
            description=step.description,
        )

    def _create_sampling_step(
        self,
        step: SamplingDicomModel,
        steps_directory: StepsDirectory,
    ) -> BaseSampling:
        """Create sampling step from sampling model.

        Will create any needed parent specimens.

        Parameters
        ----------
        step: SamplingDicomModel
            Sampling model to create step from.
        steps_directory: StepsDirectory
            DICOM preparation steps ordered by specimen identifier.

        Returns
        -------
        BaseSampling
            Sampling step created from model.
        """
        parent, sampling_constraints = self._get_parent_for_sampling(
            step, steps_directory
        )
        if isinstance(parent, Sample):
            # If Sample create sampling with constraint
            return parent.sample(
                method=step.method,
                date_time=step.date_time,
                description=step.description,
                sampling_constraints=sampling_constraints,
                location=step.sampling_location,
            )
        if isinstance(parent, Specimen):
            # Extracted specimen can not have constraint
            return parent.sample(
                method=step.method,
                date_time=step.date_time,
                description=step.description,
                location=step.sampling_location,
            )
        raise ValueError(f"Unknown parent type {type(parent)}.")

    def _get_parent_for_sampling(
        self,
        sampling: SamplingDicomModel,
        steps_directory: StepsDirectory,
    ) -> Tuple[BaseSpecimen, Optional[Sequence[BaseSampling]]]:
        """Get parent specimen and any sampling constraints for sampling.

        Creates any needed parent specimens.

        Parameters
        ----------
        sampling: SamplingDicomModel
            Sampling to get parent for.
        steps_directory: StepsDirectory
            DICOM preparation steps ordered by specimen identifier.

        Returns
        -------
        Tuple[BaseSpecimen, Optional[Sequence[BaseSampling]]]
            Parent specimen and any sampling constraints.
        """
        parent_identifier = sampling.parent_identifier
        sampling_constraints: Optional[Sequence[BaseSampling]] = None
        if parent_identifier not in self._parsed_specimens:
            # Parent does not exist, create it.
            parent = self._create_specimen(
                parent_identifier,
                sampling.parent_specimen_type,
                steps_directory,
            )
            if isinstance(parent, Sample):
                sampling_constraints = parent.sampled_from
            self._parsed_specimens[parent_identifier] = parent
            return parent, sampling_constraints

        # Parent already exists. Check that parents preparation steps match and
        # add the new sampling tree to the parent.
        parent = self._parsed_specimens[parent_identifier]
        parsed_parent = self._parse_preparation_steps_for_specimen(
            parent_identifier, steps_directory
        )
        existing_parent_steps_without_samplings = [
            parent_step
            for parent_step in parent.steps
            if not isinstance(parent_step, BaseSampling)
        ]
        if (
            not parsed_parent.preparation_steps
            == existing_parent_steps_without_samplings
        ):
            raise ValueError(
                f"Specimen {parent_identifier} already exists with different steps."
            )
        if (
            isinstance(parent, Sample)
            and parsed_parent.sampling is not None
            and parsed_parent.sampling not in parent.sampled_from_list
        ):
            # Parsed parent adds a new sampling branch
            parent.sampled_from_list.append(parsed_parent.sampling)
            sampling_constraints = [parsed_parent.sampling]
        return parent, sampling_constraints

    def _create_unknown_sampling(
        self,
        identifier: SpecimenIdentifier,
        steps_directory: StepsDirectory,
    ) -> Optional[BaseSampling]:
        """Create unknown sampling from previous specimen.

        Creates any needed parent specimens.

        Parameters
        ----------
        identifier: SpecimenIdentifier
            Identifier of the specimen the sampling belongs to.
        steps_directory: StepsDirectory
            DICOM preparation steps ordered by specimen identifier.

        Returns
        -------
        Optional[BaseSampling]
            Unknown sampling created from previous specimen.
        """

        previous_specimen = self._get_previous_specimen(identifier, steps_directory)
        if previous_specimen is None:
            return None
        sampling = previous_specimen.sample()
        return sampling

    def _get_previous_specimen(
        self,
        identifier: SpecimenIdentifier,
        steps_directory: StepsDirectory,
    ) -> Optional[Union[Specimen, Sample]]:
        """Get previous described specimen.

        Will create any needed parent specimens.

        Parameters
        ----------
        identifier: SpecimenIdentifier
            Identifier of the specimen to get previous specimen for.
        steps_directory: StepsDirectory
            DICOM preparation steps ordered by specimen identifier.

        Returns
        -------
        Optional[Union[Specimen, Sample]]
            Previous specimen, if any.
        """
        previous_specimen_identifier = steps_directory.get_previous_specimen_identifier(
            identifier
        )
        if previous_specimen_identifier is None:
            return None
        return self._create_specimen(
            previous_specimen_identifier,
            None,
            steps_directory,
        )

    def _create_specimen(
        self,
        identifier: SpecimenIdentifier,
        specimen_type: Optional[AnatomicPathologySpecimenTypesCode],
        steps_directory: StepsDirectory,
    ) -> Union[Specimen, Sample]:
        """
        Create an Specimen or Sample.

        Parameters
        ----------
        identifier: SpecimenIdentifier
            The identifier of the specimen to create.
        specimen_type: AnatomicPathologySpecimenTypesCode
            The coded type of the specimen to create.
        steps_directory: StepsDirectory
            DICOM preparation steps ordered by specimen identifier.


        Returns
        ----------
        Union[Specimen, Sample]
            Created Specimen, if the specimen has no parents, or Specimen.

        """
        logging.debug(f"Creating specimen with identifier {identifier}")
        parsed_specimen = self._parse_preparation_steps_for_specimen(
            identifier, steps_directory
        )
        if specimen_type is None:
            specimen_type = parsed_specimen.specimen_type
        elif (
            parsed_specimen.specimen_type is not None
            and specimen_type != parsed_specimen.specimen_type
        ):
            raise ValueError(
                f"Specimen type {specimen_type} and step specimen type "
                f"{parsed_specimen.specimen_type} do not match."
            )
        if parsed_specimen.sampling is None:
            if len(parsed_specimen.preparation_steps) > 0 and isinstance(
                parsed_specimen.preparation_steps[0], Collection
            ):
                collection_step = parsed_specimen.preparation_steps.pop(0)
                assert isinstance(collection_step, Collection)
            else:
                collection_step = None
            return Specimen(
                identifier=identifier.simplify(),
                extraction_step=collection_step,
                type=specimen_type,
                steps=parsed_specimen.preparation_steps,
                container=parsed_specimen.container,
            )
        return Sample(
            identifier=identifier.simplify(),
            type=specimen_type,
            sampled_from=[parsed_specimen.sampling]
            if parsed_specimen.sampling is not None
            else [],
            steps=parsed_specimen.preparation_steps,
            container=parsed_specimen.container,
        )



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""A target for writing WSI DICOM files to disk."""

from collections import defaultdict
from pathlib import Path
from typing import Callable, Dict, Iterable, List, Optional, Sequence, Set, Tuple, Union

from pydicom.uid import UID
from pydicom.valuerep import MAX_VALUE_LEN

from wsidicom.codec import Encoder
from wsidicom.codec import Settings as EncoderSettings
from wsidicom.file.io import (
    OffsetTableType,
    WsiDicomReader,
    WsiDicomWriter,
)
from wsidicom.file.wsidicom_file_image_data import WsiDicomFileImageData
from wsidicom.geometry import Size, SizeMm
from wsidicom.group import Group, Level
from wsidicom.instance import ImageData, WsiInstance
from wsidicom.series import Labels, Overviews, Pyramid, Pyramids
from wsidicom.target import Target


class WsiDicomFileTarget(Target):
    """Target for writing WSI DICOM instances to disk."""

    def __init__(
        self,
        output_path: Path,
        uid_generator: Callable[..., UID],
        workers: int,
        chunk_size: int,
        offset_table: Optional[OffsetTableType] = None,
        include_pyramids: Optional[Sequence[int]] = None,
        include_levels: Optional[Sequence[int]] = None,
        add_missing_levels: bool = False,
        transcoding: Optional[Union[EncoderSettings, Encoder]] = None,
    ):
        """
        Create a WsiDicomFileTarget.

        Parameters
        ----------
        output_path: Path
            Folder path to save files to.
        uid_generator: Callable[..., UID]
            Uid generator to use.
        workers: int
            Maximum number of thread workers to use.
        chunk_size: int
            Chunk size (number of tiles) to process at a time. Actual chunk
            size also depends on minimun_chunk_size from image_data.
        offset_table: OffsetTableType
            Offset table to use.
        include_pyramids: Optional[Sequence[int]] = None
            Optional list indices (in present pyramids) to include.
        include_levels: Optional[Sequence[int]] = None
            Optional list indices (in all pyramids) to include, e.g. [0, 1]
            includes the two lowest levels. Negative indicies can be used,
            e.g. [-1, -2] includes the two highest levels.
        add_missing_levels: bool
            If to add missing dyadic levels up to the single tile level.
        transcoding: Optional[Union[EncoderSettings, Encoder]] = None,
            Optional settings or encoder for transcoding image data. If None, image data
            will be copied as is.
        """
        self._output_path = output_path
        self._offset_table = offset_table
        self._filepaths: List[Path] = []
        self._opened_files: List[WsiDicomReader] = []
        super().__init__(
            uid_generator,
            workers,
            chunk_size,
            include_pyramids,
            include_levels,
            add_missing_levels,
            transcoding,
        )

    @property
    def filepaths(self) -> List[Path]:
        """Return filepaths for created files."""
        return self._filepaths

    def save_pyramids(self, pyramids: Pyramids):
        if self._include_pyramids is not None:
            pyramids_to_save = [pyramids[index] for index in self._include_pyramids]
        else:
            pyramids_to_save = pyramids
        for pyramid in pyramids_to_save:
            self._save_pyramid(pyramid)

    def save_labels(self, labels: Labels):
        """Save labels to target."""
        for label in labels:
            self._save_group(label, 1)

    def save_overviews(self, overviews: Overviews):
        """Save overviews to target."""
        for overview in overviews:
            self._save_group(overview, 1)

    def close(self) -> None:
        """Close any opened level files."""
        for file in self._opened_files:
            file.close()

    def _save_pyramid(self, pyramid: Pyramid):
        """Save pyramid to target."""
        # Collection of new pyramid levels.
        new_levels: List[Level] = []
        highest_level_in_file = pyramid.pyramid_indices[-1]
        lowest_single_tile_level = pyramid.lowest_single_tile_level
        highest_level = max(highest_level_in_file, lowest_single_tile_level)
        for pyramid_level in range(highest_level + 1):
            if not self._is_included_level(
                pyramid_level,
                pyramid.pyramid_indices,
                self._add_missing_levels,
                self._include_levels,
            ):
                continue
            if pyramid_level in pyramid.pyramid_indices:
                level = pyramid.get(pyramid_level)
                self._save_group(level, 1)
            elif self._add_missing_levels:
                # Create scaled level from closest level, prefer from original levels
                closest_level = pyramid.get_closest_by_level(pyramid_level)
                closest_new_level = next(
                    (
                        level
                        for level in new_levels
                        if level.level < pyramid_level
                        and level.level > closest_level.level
                    ),
                    None,
                )
                if closest_new_level is not None:
                    closest_level = closest_new_level
                scale = int(2 ** (pyramid_level - closest_level.level))
                new_level = self._save_and_open_level(
                    closest_level,
                    pyramid.pixel_spacing,
                    scale,
                )
                new_levels.append(new_level)

    def _save_and_open_level(
        self,
        level: Level,
        base_pixel_spacing: SizeMm,
        scale: int,
    ) -> Level:
        """Save level and return a new level from the created files."""
        filepaths = self._save_group(level, scale)
        instances = self._open_files(filepaths)
        return Level(instances, base_pixel_spacing)

    def _save_group(
        self,
        group: Group,
        scale: int,
    ) -> List[Path]:
        """Save group to target."""
        if not isinstance(scale, int) or scale < 1:
            raise ValueError(f"Scale must be positive integer, got {scale}.")
        filepaths: List[Path] = []
        for instances in self._group_instances_to_file(group):
            uid = self._uid_generator()
            filepath = self._output_path.joinpath(uid + ".dcm")

            image_data_list = self._list_image_data(instances)
            focal_planes, optical_paths, tiled_size = self._get_frame_information(
                image_data_list
            )

            dataset = instances[0].dataset.as_tiled_full(
                focal_planes, optical_paths, tiled_size, scale
            )
            if self._transcoder is not None:
                if (
                    self._transcoder.bits != instances[0].image_data.bits
                    or self._transcoder.samples_per_pixel
                    != instances[0].image_data.samples_per_pixel
                ):
                    raise ValueError(
                        "Transcode settings must match image data bits and photometric interpretation."
                    )
                transfer_syntax = self._transcoder.transfer_syntax
                dataset.PhotometricInterpretation = (
                    self._transcoder.photometric_interpretation
                )
                if self._transcoder.lossy_method:
                    dataset.LossyImageCompression = "01"
                    ratios = dataset.get_multi_value("LossyImageCompressionRatio")
                    # Reserve space for new ratio
                    ratios.append(" " * MAX_VALUE_LEN["DS"])
                    methods = dataset.get_multi_value("LossyImageCompressionMethod")
                    methods.append(self._transcoder.lossy_method.value)
                    dataset.LossyImageCompressionRatio = ratios
                    dataset.LossyImageCompressionMethod = methods

            else:
                transfer_syntax = instances[0].image_data.transfer_syntax
            if self._offset_table is not None:
                offset_table = self._offset_table
            elif transfer_syntax.is_encapsulated:
                offset_table = OffsetTableType.BASIC
            else:
                offset_table = OffsetTableType.NONE
            with WsiDicomWriter.open(filepath, transfer_syntax, offset_table) as writer:
                writer.write(
                    uid,
                    dataset,
                    image_data_list,
                    self._workers,
                    self._chunk_size,
                    self._instance_number,
                    scale,
                    self._transcoder,
                )
            filepaths.append(filepath)
            self._instance_number += 1
        self._filepaths.extend(filepaths)
        return filepaths

    def _open_files(self, filepaths: Iterable[Path]) -> List[WsiInstance]:
        readers = [WsiDicomReader.open(filepath) for filepath in filepaths]
        self._opened_files.extend(readers)
        return [
            WsiInstance(
                [reader.dataset for reader in readers], WsiDicomFileImageData(readers)
            )
        ]

    @staticmethod
    def _group_instances_to_file(group: Group) -> List[List[WsiInstance]]:
        """
        Group instances by properties that can't differ in a DICOM-file.

        Returns
        -------
        List[List[WsiInstance]]
            Instances grouped by common properties.
        """
        groups: Dict[
            Tuple[str, UID, bool, Optional[int], Optional[float], str],
            List[WsiInstance],
        ] = defaultdict(list)

        for instance in group.instances.values():
            groups[
                instance.image_data.photometric_interpretation,
                instance.image_data.transfer_syntax,
                instance.ext_depth_of_field,
                instance.ext_depth_of_field_planes,
                instance.ext_depth_of_field_plane_distance,
                instance.focus_method,
            ].append(instance)
        return list(groups.values())

    @staticmethod
    def _list_image_data(
        instances: Iterable[WsiInstance],
    ) -> Dict[Tuple[str, float], ImageData]:
        """
        Sort ImageData in instances by optical path and focal plane.

        Parameters
        ----------
        instances: Iterable[WsiInstance]
            List of instances with optical paths and focal planes to list and
            sort.

        Returns
        -------
        Dict[Tuple[str, float], ImageData]:
            ImageData sorted by optical path and focal plane.
        """
        output: Dict[Tuple[str, float], ImageData] = {}
        for instance in instances:
            for optical_path in instance.optical_paths:
                for z in sorted(instance.focal_planes):
                    if (optical_path, z) not in output:
                        output[optical_path, z] = instance.image_data
        return output

    @staticmethod
    def _get_frame_information(
        data: Dict[Tuple[str, float], ImageData]
    ) -> Tuple[List[float], List[str], Size]:
        """Return optical_paths, focal planes, and tiled size."""
        focal_planes_by_optical_path: Dict[str, Set[float]] = defaultdict(set)
        all_focal_planes: Set[float] = set()
        tiled_sizes: Set[Size] = set()
        for (optical_path, focal_plane), image_data in data.items():
            focal_planes_by_optical_path[optical_path].add(focal_plane)
            all_focal_planes.add(focal_plane)
            tiled_sizes.add(image_data.tiled_size)

        focal_planes_sparse_by_optical_path = any(
            optical_path_focal_planes != all_focal_planes
            for optical_path_focal_planes in focal_planes_by_optical_path.values()
        )
        if focal_planes_sparse_by_optical_path:
            raise ValueError("Each optical path must have the same focal planes.")

        if len(tiled_sizes) != 1:
            raise ValueError(f"Expected only one tiled size, found {len(tiled_sizes)}.")
        tiled_size = list(tiled_sizes)[0]
        return (
            sorted(list(all_focal_planes)),
            sorted(list(focal_planes_by_optical_path.keys())),
            tiled_size,
        )



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""A source for reading WSI DICOM files."""

import io
import logging
from collections import defaultdict
from pathlib import Path
from typing import BinaryIO, Dict, Iterable, List, Optional, Union

from pydicom import dcmread
from pydicom.errors import InvalidDicomError
from pydicom.fileset import FileSet
from pydicom.uid import UID, MediaStorageDirectoryStorage

from wsidicom.errors import (
    WsiDicomNotFoundError,
    WsiDicomNotSupportedError,
)
from wsidicom.file.io import WsiDicomReader
from wsidicom.file.io.wsidicom_io import WsiDicomIO
from wsidicom.file.wsidicom_file_image_data import WsiDicomFileImageData
from wsidicom.geometry import Size
from wsidicom.graphical_annotations import AnnotationInstance
from wsidicom.instance import ImageType, TileType, WsiDataset, WsiInstance
from wsidicom.source import Source
from wsidicom.uid import ANN_SOP_CLASS_UID, WSI_SOP_CLASS_UID, SlideUids


class WsiDicomFileSource(Source):
    """Source reading WSI DICOM file instances."""

    def __init__(
        self,
        files: Union[str, Path, BinaryIO, Iterable[Union[str, Path, BinaryIO]]],
    ) -> None:
        """Create a WsiDicomFileSource.

        Parameters
        ----------
        files: Union[str, Path, BinaryIO, Iterable[Union[str, Path, BinaryIO]]],
            Files to open. Can be a path or stream for a single file, a list of paths or
            streams for multiple files, or a path to a folder containing files.
        """
        self._levels: List[WsiDicomReader] = []
        self._labels: List[WsiDicomReader] = []
        self._overviews: List[WsiDicomReader] = []
        self._annotations: List[WsiDicomIO] = []
        for file in self._list_input_files(files):
            try:
                stream = self._open_file(file)
                sop_class_uid = self._get_sop_class_uid(stream)
                if sop_class_uid == WSI_SOP_CLASS_UID:
                    try:
                        reader = WsiDicomReader(stream)
                        if reader.image_type == ImageType.VOLUME:
                            self._levels.append(reader)
                        elif reader.image_type == ImageType.LABEL:
                            self._labels.append(reader)
                        elif reader.image_type == ImageType.OVERVIEW:
                            self._overviews.append(reader)
                    except WsiDicomNotSupportedError:
                        logging.info(f"Non-supported file {stream}.")
                        if stream.filepath is not None:
                            stream.close()
                elif sop_class_uid == ANN_SOP_CLASS_UID:
                    self._annotations.append(stream)
                elif stream.filepath is not None:
                    logging.debug(
                        f"Non-supported SOP class {sop_class_uid} "
                        f"for file {stream}."
                    )
                    # File was opened but not supported SOP class.
                    stream.close()
            except Exception:
                logging.error(
                    f"Failed to open file {file} due to exception.", exc_info=True
                )
        if len(self._levels) == 0:
            raise WsiDicomNotFoundError("Level files", str(files))
        self._base_dataset = self._get_base_dataset(self._levels)
        self._slide_uids = self._base_dataset.uids.slide
        self._base_tile_size = self._base_dataset.tile_size

    @property
    def base_dataset(self) -> WsiDataset:
        """The dataset of the base level instance."""
        return self._base_dataset

    @property
    def level_instances(self) -> Iterable[WsiInstance]:
        """The level instances parsed from the source."""
        return self._create_instances(
            self._levels, self._slide_uids, self._base_tile_size
        )

    @property
    def label_instances(self) -> Iterable[WsiInstance]:
        """The label instances parsed from the source."""
        return self._create_instances(self._labels, self._slide_uids)

    @property
    def overview_instances(self) -> Iterable[WsiInstance]:
        """The overview instances parsed from the source."""
        return self._create_instances(self._overviews, self._slide_uids)

    @property
    def annotation_instances(self) -> Iterable[AnnotationInstance]:
        """The annotation instances parsed from the source."""
        return [
            AnnotationInstance.open_dataset(file.read_dataset())
            for file in self._annotations
        ]

    def close(self) -> None:
        """Close all opened readers in the source. Does not close provided streams."""
        for reader in self.readers:
            reader.close()

    @property
    def readers(self) -> List[WsiDicomReader]:
        """Return the readers in the source."""
        reader_lists: List[List[WsiDicomReader]] = [
            self._levels,
            self._labels,
            self._overviews,
        ]
        return [reader for reader_list in reader_lists for reader in reader_list]

    @property
    def files(self) -> Optional[List[Path]]:
        """Return the files in the source, if any."""
        if all(reader.filepath is None for reader in self.readers):
            return None
        return [
            reader.filepath for reader in self.readers if reader.filepath is not None
        ]

    @property
    def is_ready_for_viewing(self) -> Optional[bool]:
        """
        Returns True if files in source are formatted for fast viewing.

        Returns None if no files are in source.
        """
        if not self.contains_levels:
            return None
        files = sorted(self.readers, key=lambda file: file.frame_count)
        for file in files:
            if file.image_type is None:
                continue
            if (
                file.dataset.tile_type != TileType.SPARSE
                or file.offset_table_type is None
            ):
                return False

        return True

    @property
    def contains_levels(self) -> bool:
        """Returns true source has one level that can be read with WsiDicom."""
        return len(self._levels) > 0

    @classmethod
    def open_dicomdir(cls, path: Union[str, Path]):
        """Open a DICOMDIR file and return a WsiDicomFileSource for contained files.

        Parameters
        ----------
        path: Union[str, Path]
            Path to DICOMDIR file.

        Returns
        -------
        WsiDicomFileSource
            Source for files in DICOMDIR.
        """
        dicomdir = dcmread(path)
        if dicomdir.file_meta.MediaStorageSOPClassUID != MediaStorageDirectoryStorage:
            raise ValueError()
        fileset = FileSet(dicomdir)
        files = [file.path for file in fileset]
        return cls(files)

    @staticmethod
    def _open_file(file: Union[Path, BinaryIO]) -> WsiDicomIO:
        """Open stream if file is path. Return stream and optional filepath."""
        if isinstance(file, Path):
            return WsiDicomIO.open(file, "rb")
        return WsiDicomIO(file)

    @staticmethod
    def _list_input_files(
        files: Union[str, Path, BinaryIO, Iterable[Union[str, Path, BinaryIO]]],
    ) -> Iterable[Union[Path, BinaryIO]]:
        """List input files. Iterate directory content if directory.

        Parameters
        ----------
        files: Union[str, Path, BinaryIO, Iterable[Union[str, Path, BinaryIO]]],
            Files or directory to list.

        Returns
        -------
        Iterable[Tuple[BinaryIO, Optional[Path]]]:
            Iterable files to open.
        """

        if isinstance(files, (str, Path)):
            # Path to single file or folder with files.
            single_path = Path(files)
            if single_path.is_dir():
                return (file for file in single_path.iterdir() if file.is_file())
            if single_path.is_file():
                return [single_path]
            raise ValueError(f"File in path {single_path} was not a file or directory.")

        if isinstance(files, BinaryIO):
            # Single stream.
            return [files]

        # Multiple paths or streams.
        return (
            Path(file) if isinstance(file, str) else file
            for file in files
            if isinstance(file, io.IOBase)
            or (isinstance(file, (str, Path)) and Path(file).is_file())
        )

    @staticmethod
    def _get_base_dataset(files: Iterable[WsiDicomReader]) -> WsiDataset:
        """Return file with largest image (width) from list of files.

        Parameters
        ----------
        files: Iterable[WsiDicomReader]
           List of files.

        Returns
        -------
        WsiDataset
            Base layer dataset.
        """
        return next(
            file.dataset
            for file in sorted(
                files, reverse=True, key=lambda file: file.dataset.image_size.width
            )
        )

    @staticmethod
    def _get_sop_class_uid(stream: WsiDicomIO) -> Optional[UID]:
        """Return the SOP class UID from file metadata or None if invalid DICOM."""
        try:
            return stream.read_media_storage_sop_class_uid()
        except InvalidDicomError as exception:
            logging.debug(
                f"Failed to parse DICOM file metadata for file {stream}, not DICOM? "
                f"Got exception {exception}"
            )
            return None

    @classmethod
    def _create_instances(
        cls,
        files: Iterable[WsiDicomReader],
        series_uids: SlideUids,
        series_tile_size: Optional[Size] = None,
    ) -> Iterable["WsiInstance"]:
        """
        Create instances from Dicom files.

        Only files with matching series uid and tile size, if defined, are used. Other
        files are closed.

        Parameters
        ----------
        files: Sequence[WsiDicomReader]
            Files to create instances from.
        series_uids: SlideUids
            Uid to match against.
        series_tile_size: Optional[Size]
            Tile size to match against (for level instances).

        Returns
        -------
        Iterable[WsiInstancece]
            Iterable of created instances.
        """
        filtered_files = cls._filter_files(files, series_uids, series_tile_size)
        files_grouped_by_instance = cls._group_files(filtered_files)
        return (
            WsiInstance(
                [file.dataset for file in instance_files],
                WsiDicomFileImageData(instance_files),
            )
            for instance_files in files_grouped_by_instance.values()
        )

    @staticmethod
    def _filter_files(
        files: Iterable[WsiDicomReader],
        series_uids: SlideUids,
        series_tile_size: Optional[Size] = None,
    ) -> Iterable[WsiDicomReader]:
        """
        Filter list of wsi dicom files on uids and tile size if defined.

        Parameters
        ----------
        files: Iterable[WsiDicomReader]
            Wsi files to filter.
        series_uids: Uids
            Uids to check against.
        series_tile_size: Optional[Size] = None
            Tile size to check against.

        Returns
        -------
        List[WsiDicomReader]
            List of matching wsi dicom files.
        """
        for file in files:
            if file.dataset.matches_series(series_uids, series_tile_size):
                yield file
            else:
                logging.warning(
                    f"{file.filepath} with uids {file.uids.slide} "
                    f"did not match series with {series_uids} "
                    f"and tile size {series_tile_size}"
                )
                file.close()

    @staticmethod
    def _group_files(
        files: Iterable[WsiDicomReader],
    ) -> Dict[str, List[WsiDicomReader]]:
        """
        Return files grouped by instance identifier (instances).

        Parameters
        ----------
        files: Iterable[WsiDicomReader]
            Files to group into instances

        Returns
        -------
        Dict[str, List[WsiDicomReader]]
            Files grouped by instance, with instance identifier as key.
        """
        grouped_files: Dict[str, List[WsiDicomReader]] = defaultdict(list)
        for file in files:
            grouped_files[file.uids.identifier].append(file)
        return grouped_files



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

__all__ = [
    "settings",
    "Annotation",
    "AnnotationGroup",
    "AnnotationInstance",
    "Measurement",
    "Point",
    "PointAnnotationGroup",
    "Polygon",
    "PolygonAnnotationGroup",
    "Polyline",
    "PolylineAnnotationGroup",
    "ImageData",
    "WsiDataset",
    "WsiInstance",
    "Labels",
    "Pyramid",
    "Overviews",
    "WsiDicomWebClient",
    "WsiDicom",
]
from wsidicom.config import settings
from wsidicom.graphical_annotations import (
    Annotation,
    AnnotationGroup,
    AnnotationInstance,
    Measurement,
    Point,
    PointAnnotationGroup,
    Polygon,
    PolygonAnnotationGroup,
    Polyline,
    PolylineAnnotationGroup,
)
from wsidicom.instance import ImageData, WsiDataset, WsiInstance
from wsidicom.series import Labels, Overviews, Pyramid
from wsidicom.web import WsiDicomWebClient
from wsidicom.wsidicom import WsiDicom

__version__ = "0.19.0"



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""DICOM schema for complete WsiMetadata model."""

from dataclasses import replace
from typing import Sequence, Type

from marshmallow import fields
from pydicom import Dataset
from pydicom.uid import VLWholeSlideMicroscopyImageStorage

from wsidicom.instance.dataset import ImageType
from wsidicom.metadata.equipment import Equipment
from wsidicom.metadata.image import Image
from wsidicom.metadata.label import Label
from wsidicom.metadata.optical_path import OpticalPath
from wsidicom.metadata.patient import Patient
from wsidicom.metadata.schema.dicom.equipment import EquipmentDicomSchema
from wsidicom.metadata.schema.dicom.fields import (
    DefaultingListDicomField,
    DefaultingListTagDicomField,
    DefaultingTagDicomField,
    FlattenOnDumpNestedDicomField,
    UidDatasetDicomField,
    UidDicomField,
)
from wsidicom.metadata.schema.dicom.image import ImageDicomSchema
from wsidicom.metadata.schema.dicom.label import LabelDicomSchema
from wsidicom.metadata.schema.dicom.optical_path import OpticalPathDicomSchema
from wsidicom.metadata.schema.dicom.patient import PatientDicomSchema
from wsidicom.metadata.schema.dicom.schema import DicomSchema
from wsidicom.metadata.schema.dicom.series import SeriesDicomSchema
from wsidicom.metadata.schema.dicom.slide import SlideDicomSchema
from wsidicom.metadata.schema.dicom.study import StudyDicomSchema
from wsidicom.metadata.series import Series
from wsidicom.metadata.slide import Slide
from wsidicom.metadata.study import Study
from wsidicom.metadata.wsi import WsiMetadata


class WsiMetadataDicomSchema(DicomSchema[WsiMetadata]):
    study = FlattenOnDumpNestedDicomField(
        StudyDicomSchema(), dump_default=Study(), load_default=Study()
    )
    series = FlattenOnDumpNestedDicomField(
        SeriesDicomSchema(), dump_default=Series(), load_default=Series()
    )
    patient = FlattenOnDumpNestedDicomField(
        PatientDicomSchema(), dump_default=Patient(), load_default=Patient()
    )
    equipment = FlattenOnDumpNestedDicomField(
        EquipmentDicomSchema(), dump_default=Equipment(), load_default=Equipment()
    )
    optical_paths = DefaultingListDicomField(
        FlattenOnDumpNestedDicomField(OpticalPathDicomSchema()),
        data_key="OpticalPathSequence",
        dump_default=[OpticalPath()],
        load_default=[],
    )
    slide = FlattenOnDumpNestedDicomField(
        SlideDicomSchema(), dump_default=Slide(), load_default=Slide()
    )
    label = FlattenOnDumpNestedDicomField(
        LabelDicomSchema(), dump_default=Label(), load_default=Label()
    )
    image = FlattenOnDumpNestedDicomField(
        ImageDicomSchema(), dump_default=Image(), load_default=Image()
    )
    frame_of_reference_uid = DefaultingTagDicomField(
        UidDicomField(),
        allow_none=True,
        data_key="FrameOfReferenceUID",
        tag="default_frame_of_reference_uid",
    )
    dimension_organization_uids = DefaultingListTagDicomField(
        UidDatasetDicomField(data_key="DimensionOrganizationUID"),
        tag="default_dimension_organization_uids",
        data_key="DimensionOrganizationSequence",
    )
    sop_class_uid = fields.Constant(
        VLWholeSlideMicroscopyImageStorage, dump_only=True, data_key="SOPClassUID"
    )
    modality = fields.Constant("SM", dump_only=True, data_key="Modality")
    position_reference_indicator = fields.Constant(
        "SLIDE_CORNER", dump_only=True, data_key="PositionReferenceIndicator"
    )
    volumetric_properties = fields.Constant(
        "VOLUME", dump_only=True, data_key="VolumetricProperties"
    )
    acquisition_context = fields.Constant(
        [], data_key="AcquisitionContextSequence", dump_only=True
    )

    @property
    def load_type(self) -> Type[WsiMetadata]:
        return WsiMetadata

    @classmethod
    def from_datasets(cls, datasets: Sequence[Dataset]) -> WsiMetadata:
        label_dataset = next(
            (
                dataset
                for dataset in datasets
                if dataset.ImageType[2] == ImageType.LABEL.value
            ),
            None,
        )
        overview_dataset = next(
            (
                dataset
                for dataset in datasets
                if dataset.ImageType[2] == ImageType.OVERVIEW.value
            ),
            None,
        )
        volume_dataset = next(
            dataset
            for dataset in datasets
            if dataset.ImageType[2] == ImageType.VOLUME.value
        )
        label_dicom_schema = LabelDicomSchema()
        metadata = WsiMetadataDicomSchema().load(volume_dataset)
        if label_dataset is None:
            label_label = None
        else:
            label_label = label_dicom_schema.load(label_dataset)
        if overview_dataset is None:
            overview_label = None
        else:
            overview_label = label_dicom_schema.load(overview_dataset)
        assert metadata.label is not None
        merged_label = Label.merge_image_types(
            metadata.label, label_label, overview_label
        )
        return replace(metadata, label=merged_label)



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Models for use during DICOM specimen serialization."""

import datetime
from abc import ABCMeta
from dataclasses import dataclass, field
from typing import (
    List,
    Optional,
    Sequence,
    Union,
)

from pydicom.sr.coding import Code
from pydicom.uid import UID

from wsidicom.conceptcode import (
    AnatomicPathologySpecimenTypesCode,
    ContainerTypeCode,
    SpecimenCollectionProcedureCode,
    SpecimenEmbeddingMediaCode,
    SpecimenFixativesCode,
    SpecimenPreparationStepsCode,
    SpecimenSamplingProcedureCode,
    SpecimenStainsCode,
)
from wsidicom.metadata.sample import (
    BaseSpecimen,
    Collection,
    Embedding,
    Fixation,
    IssuerOfIdentifier,
    Measurement,
    PreparationStep,
    Processing,
    Receiving,
    SampleLocalization,
    Sampling,
    SamplingLocation,
    SpecimenIdentifier,
    Staining,
    Storage,
    UnknownSampling,
)


class SpecimenPreparationStepDicomModel(metaclass=ABCMeta):
    def __init__(
        self,
        identifier: str,
        issuer_of_identifier: Optional[str] = None,
        date_time: Optional[datetime.datetime] = None,
        description: Optional[str] = None,
        fixative: Optional[SpecimenFixativesCode] = None,
        embedding: Optional[SpecimenEmbeddingMediaCode] = None,
        processing: Optional[SpecimenPreparationStepsCode] = None,
        specimen_type: Optional[AnatomicPathologySpecimenTypesCode] = None,
        container: Optional[ContainerTypeCode] = None,
    ):
        self.identifier = identifier
        self.issuer_of_identifier = issuer_of_identifier
        self.date_time = date_time
        self.description = description
        self.fixative = fixative
        self.embedding = embedding
        self.processing = processing
        self.specimen_type = specimen_type
        self.container = container

    @classmethod
    def from_step(
        cls,
        step: PreparationStep,
        specimen: BaseSpecimen,
    ) -> Optional["SpecimenPreparationStepDicomModel"]:
        """Return DICOM model for the step.

        Parameters
        ----------
        step: PreparationStep
            Step to convert into DICOM model.
        specimen: BaseSpecimen
            Specimen that was processed.

        Returns
        -------
        Optional[SpecimenPreparationStepDicomModel]
            DICOM model for the step, or None if no step should be produced.
        """
        if isinstance(step, Sampling):
            if step.specimen.type is None:
                return None
            return SamplingDicomModel.from_step(step, specimen)
        if isinstance(step, UnknownSampling):
            return None
        if isinstance(step, Collection):
            return CollectionDicomModel.from_step(step, specimen)
        if isinstance(step, (Processing, Embedding, Fixation)):
            return ProcessingDicomModel.from_step(step, specimen)
        if isinstance(step, Staining):
            return StainingDicomModel.from_step(step, specimen)
        if isinstance(step, Receiving):
            return ReceivingDicomModel.from_step(step, specimen)
        if isinstance(step, Storage):
            return StorageDicomModel.from_step(step, specimen)

        raise NotImplementedError(f"Unknown preparation step type {type(step)}.")

    @property
    def specimen_identifier(self) -> SpecimenIdentifier:
        if self.issuer_of_identifier is None:
            issuer = None
        else:
            issuer = IssuerOfIdentifier.from_hl7v2(self.issuer_of_identifier)
        return SpecimenIdentifier(
            self.identifier,
            issuer,
        )


@dataclass
class SamplingDicomModel(SpecimenPreparationStepDicomModel):
    identifier: str
    method: SpecimenSamplingProcedureCode
    parent_specimen_identifier: str
    parent_specimen_type: AnatomicPathologySpecimenTypesCode
    issuer_of_identifier: Optional[str] = None
    issuer_of_parent_specimen_identifier: Optional[str] = None
    date_time: Optional[datetime.datetime] = None
    description: Optional[str] = None
    fixative: Optional[SpecimenFixativesCode] = None
    embedding: Optional[SpecimenEmbeddingMediaCode] = None
    processing: Optional[SpecimenPreparationStepsCode] = None
    specimen_type: Optional[AnatomicPathologySpecimenTypesCode] = None
    container: Optional[ContainerTypeCode] = None
    location_reference: Optional[str] = None
    location_description: Optional[str] = None
    location_x: Optional[Measurement] = None
    location_y: Optional[Measurement] = None
    location_z: Optional[Measurement] = None

    def __post_init__(self):
        super().__init__(
            identifier=self.identifier,
            issuer_of_identifier=self.issuer_of_identifier,
            date_time=self.date_time,
            description=self.description,
            fixative=self.fixative,
            embedding=self.embedding,
            processing=self.processing,
            specimen_type=self.specimen_type,
            container=self.container,
        )

    @classmethod
    def from_step(
        cls,
        sampling: Sampling,
        specimen: BaseSpecimen,
    ) -> "SamplingDicomModel":
        """Return Dicom dataset for the step.

        Parameters
        ----------
        sampling: Sampling
            Step to convert into dataset.
        specimen_identifier: Union[str, SpecimenIdentifier]:
            Identifier for the specimen that was processed.

        Parameters
        ----------
        SpecimenPreparationStep:
            Dicom dataset describing the processing step.

        """
        if sampling.specimen.type is None:
            raise ValueError("Sampled specimen must have a specimen type.")
        identifier, issuer = SpecimenIdentifier.get_string_identifier_and_issuer(
            specimen.identifier
        )
        (
            parent_identifier,
            parent_issuer,
        ) = SpecimenIdentifier.get_string_identifier_and_issuer(
            sampling.specimen.identifier
        )
        return cls(
            identifier=identifier,
            issuer_of_identifier=issuer,
            date_time=sampling.date_time,
            description=sampling.description,
            method=sampling.method,
            parent_specimen_identifier=parent_identifier,
            issuer_of_parent_specimen_identifier=parent_issuer,
            parent_specimen_type=sampling.specimen.type,
            location_reference=sampling.location.reference
            if sampling.location is not None
            else None,
            location_description=sampling.location.description
            if sampling.location is not None
            else None,
            location_x=sampling.location.x if sampling.location is not None else None,
            location_y=sampling.location.y if sampling.location is not None else None,
            location_z=sampling.location.z if sampling.location is not None else None,
            container=specimen.container,
            specimen_type=specimen.type,
        )

    @property
    def parent_identifier(self) -> SpecimenIdentifier:
        if self.issuer_of_parent_specimen_identifier is None:
            issuer = None
        else:
            issuer = IssuerOfIdentifier.from_hl7v2(
                self.issuer_of_parent_specimen_identifier
            )
        return SpecimenIdentifier(
            self.parent_specimen_identifier,
            issuer,
        )

    @property
    def sampling_location(self) -> Optional[SamplingLocation]:
        if any(
            item is not None
            for item in (
                self.location_reference,
                self.location_description,
                self.location_x,
                self.location_y,
                self.location_z,
            )
        ):
            return SamplingLocation(
                reference=self.location_reference,
                description=self.location_description,
                x=self.location_x,
                y=self.location_y,
                z=self.location_z,
            )
        return None


@dataclass
class CollectionDicomModel(SpecimenPreparationStepDicomModel):
    identifier: str
    method: SpecimenCollectionProcedureCode
    issuer_of_identifier: Optional[str] = None
    date_time: Optional[datetime.datetime] = None
    description: Optional[str] = None
    fixative: Optional[SpecimenFixativesCode] = None
    embedding: Optional[SpecimenEmbeddingMediaCode] = None
    processing: Optional[SpecimenPreparationStepsCode] = None
    specimen_type: Optional[AnatomicPathologySpecimenTypesCode] = None
    container: Optional[ContainerTypeCode] = None

    def __post_init__(self):
        super().__init__(
            identifier=self.identifier,
            issuer_of_identifier=self.issuer_of_identifier,
            date_time=self.date_time,
            description=self.description,
            fixative=self.fixative,
            embedding=self.embedding,
            processing=self.processing,
            specimen_type=self.specimen_type,
            container=self.container,
        )

    @classmethod
    def from_step(
        cls,
        collection: Collection,
        specimen: BaseSpecimen,
    ) -> "CollectionDicomModel":
        """Return Dicom dataset for the step.

        Parameters
        ----------
        collection: Collection
            Step to convert into dataset.
        specimen_identifier: Union[str, SpecimenIdentifier]:
            Identifier for the specimen that was processed.

        Parameters
        ----------
        SpecimenPreparationStep:
            Dicom dataset describing the processing step.

        """
        identifier, issuer = SpecimenIdentifier.get_string_identifier_and_issuer(
            specimen.identifier
        )
        return cls(
            identifier=identifier,
            issuer_of_identifier=issuer,
            date_time=collection.date_time,
            description=collection.description,
            method=collection.method,
            container=specimen.container,
            specimen_type=specimen.type,
        )


@dataclass
class ReceivingDicomModel(SpecimenPreparationStepDicomModel):
    identifier: str
    issuer_of_identifier: Optional[str] = None
    date_time: Optional[datetime.datetime] = None
    description: Optional[str] = None
    fixative: Optional[SpecimenFixativesCode] = None
    embedding: Optional[SpecimenEmbeddingMediaCode] = None
    processing: Optional[SpecimenPreparationStepsCode] = None
    specimen_type: Optional[AnatomicPathologySpecimenTypesCode] = None
    container: Optional[ContainerTypeCode] = None

    @classmethod
    def from_step(
        cls,
        receiving: Receiving,
        specimen: BaseSpecimen,
    ):
        identifier, issuer = SpecimenIdentifier.get_string_identifier_and_issuer(
            specimen.identifier
        )
        return cls(
            identifier=identifier,
            issuer_of_identifier=issuer,
            date_time=receiving.date_time,
            description=receiving.description,
            container=specimen.container,
            specimen_type=specimen.type,
        )


@dataclass
class StorageDicomModel(SpecimenPreparationStepDicomModel):
    identifier: str
    issuer_of_identifier: Optional[str] = None
    date_time: Optional[datetime.datetime] = None
    description: Optional[str] = None
    fixative: Optional[SpecimenFixativesCode] = None
    embedding: Optional[SpecimenEmbeddingMediaCode] = None
    processing: Optional[SpecimenPreparationStepsCode] = None
    specimen_type: Optional[AnatomicPathologySpecimenTypesCode] = None
    container: Optional[ContainerTypeCode] = None

    @classmethod
    def from_step(
        cls,
        storage: Storage,
        specimen: BaseSpecimen,
    ):
        identifier, issuer = SpecimenIdentifier.get_string_identifier_and_issuer(
            specimen.identifier
        )
        return cls(
            identifier=identifier,
            issuer_of_identifier=issuer,
            date_time=storage.date_time,
            description=storage.description,
            container=specimen.container,
            specimen_type=specimen.type,
        )


@dataclass
class ProcessingDicomModel(SpecimenPreparationStepDicomModel):
    identifier: str
    issuer_of_identifier: Optional[str] = None
    date_time: Optional[datetime.datetime] = None
    description: Optional[str] = None
    fixative: Optional[SpecimenFixativesCode] = None
    embedding: Optional[SpecimenEmbeddingMediaCode] = None
    processing: Optional[SpecimenPreparationStepsCode] = None
    specimen_type: Optional[AnatomicPathologySpecimenTypesCode] = None
    container: Optional[ContainerTypeCode] = None

    @classmethod
    def from_step(
        cls,
        processing: Union[Processing, Embedding, Fixation],
        specimen: BaseSpecimen,
    ) -> "ProcessingDicomModel":
        """Return Dicom dataset for the step.

        Parameters
        ----------
        processing: Processing
            Step to convert into dataset.
        specimen_identifier: Union[str, SpecimenIdentifier]:
            Identifier for the specimen that was processed.

        Parameters
        ----------
        SpecimenPreparationStep:
            Dicom dataset describing the processing step.

        """
        identifier, issuer = SpecimenIdentifier.get_string_identifier_and_issuer(
            specimen.identifier
        )
        if isinstance(processing, Processing):
            method = processing.method
        else:
            method = None
        if isinstance(processing, Fixation):
            fixative = processing.fixative
        else:
            fixative = None
        if isinstance(processing, Embedding):
            embedding = processing.medium
        else:
            embedding = None

        return cls(
            identifier=identifier,
            issuer_of_identifier=issuer,
            date_time=processing.date_time,
            description=processing.description,
            fixative=fixative,
            embedding=embedding,
            processing=method,
            container=specimen.container,
            specimen_type=specimen.type,
        )


@dataclass
class StainingDicomModel(SpecimenPreparationStepDicomModel):
    identifier: str
    substances: Sequence[Union[str, SpecimenStainsCode]]
    issuer_of_identifier: Optional[str] = None
    date_time: Optional[datetime.datetime] = None
    description: Optional[str] = None
    fixative: Optional[SpecimenFixativesCode] = None
    embedding: Optional[SpecimenEmbeddingMediaCode] = None
    processing: Optional[SpecimenPreparationStepsCode] = None
    specimen_type: Optional[AnatomicPathologySpecimenTypesCode] = None
    container: Optional[ContainerTypeCode] = None

    @classmethod
    def from_step(
        cls,
        staining: Staining,
        specimen: BaseSpecimen,
    ) -> "StainingDicomModel":
        """Return Dicom dataset for the step.

        Parameters
        ----------
        specimen_identifier: Union[str, SpecimenIdentifier]:
            Identifier for the specimen that was processed.

        Parameters
        ----------
        SpecimenPreparationStep:
            Dicom dataset describing the processing step.

        """
        identifier, issuer = SpecimenIdentifier.get_string_identifier_and_issuer(
            specimen.identifier
        )
        return cls(
            identifier=identifier,
            issuer_of_identifier=issuer,
            date_time=staining.date_time,
            description=staining.description,
            substances=staining.substances,
            container=specimen.container,
            specimen_type=specimen.type,
        )


@dataclass
class SpecimenDescriptionDicomModel:
    """A sample that has been placed on a slide."""

    identifier: str
    uid: UID
    steps: List[SpecimenPreparationStepDicomModel] = field(default_factory=list)
    anatomical_sites: List[Code] = field(default_factory=list)
    issuer_of_identifier: Optional[IssuerOfIdentifier] = None
    short_description: Optional[str] = None
    detailed_description: Optional[str] = None
    localization: Optional[SampleLocalization] = None

    @property
    def specimen_identifier(self) -> SpecimenIdentifier:
        return SpecimenIdentifier(self.identifier, self.issuer_of_identifier)
